/*
 * Created on 2005-04-13
 */
package gui.controls;

import gui.dialogs.ConnectionTimetableCreator;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import world.Connection;
import world.World;

/**
 * A panel used to edit and/or remove connections from those already added to the world. 
 * @author Piotrek
 */
public final class ConnectionEditChooser extends JPanel implements ActionListener, ListSelectionListener
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param worldView A view rendering the world we are editing the connections from.
	 */
	public ConnectionEditChooser(WorldView worldView)
	{
		super(null);
		
		if (worldView == null)
			throw new NullPointerException("World view must not be null in ConnectionEditChooser.ConnectionEditChooser().");
		
		this.worldView = worldView;
		
		worldView.SetConnectionEditChooser(this);
		
		setBackground(Color.BLUE);
		
		// creating a list view
		 
		connectionsList = new JList();
		
		connectionsList.setModel(listModel);
		connectionsList.setCellRenderer(new ConnectionsCellRenderer());
		
		connectionsList.addListSelectionListener(this);
	
		connectionsList.setFixedCellWidth(160);
		connectionsList.setFixedCellHeight(40);
				
		listPane = new JScrollPane(connectionsList);
		
		add(listPane);
		
		connectionsList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		
		// adding the edit button
		
		editButton = new JButton("Edit");
		
		editButton.setEnabled(false);
		editButton.setBackground(Color.GREEN);
		editButton.addActionListener(this);
		
		add(editButton);
		
		// adding the remove button
		
		removeButton = new JButton("Remove");
		
		removeButton.setEnabled(false);
		removeButton.setBackground(Color.GREEN);
		removeButton.addActionListener(this);
		
		add(removeButton);
		
		setVisible(false);
	}
	
	/**
	 * An `edit' button.
	 * @author Piotrek
	 */
	private JButton editButton = null;
	
	/**
	 * A `remove' button.
	 * @author Piotrek
	 */
	private JButton removeButton = null;
	
	/**
	 * A view displaying a world from which the connections are being edited.
	 * @author Piotrek
	 */
	private WorldView worldView = null;
	
	/**
	 * A list control displaying the connections.
	 * @author Piotrek
	 */
	private JList connectionsList = null;
	
	/**
	 * The list model providing an API to add and remove items from the list.
	 * @author Piotrek
	 */
	private DefaultListModel listModel = new DefaultListModel();
		
	/**
	 * The scroll pane making the list scrollable.
	 * @author Piotrek
	 */
	private JScrollPane listPane;
	
	/**
	 * @author Piotrek
	 * @see JComponent#paint(java.awt.Graphics)
	 */
	public void paint(Graphics g)
	{
		super.paint(g);
		
		g.setColor(getBackground());
		g.fillRect(0, 0 , getSize().width, getSize().height);
						
		paintChildren(g);
	}
	
	/**
	 * @author Piotrek
	 * @see java.awt.Component#setBounds(int, int, int, int)
	 */
	public void setBounds(int x, int y, int width, int height)
	{
		super.setBounds(x, y, width, height);
		
		listPane.setBounds(5, 10, width - 10, height - 55);
		
		editButton.setBounds(5, height - 35, (width - 20) / 2, 25);
		removeButton.setBounds(15 + (width - 20) / 2, height - 35, (width - 20) / 2, 25);
	}
	
	/**
	 * Removes the currently selected connection from the list.
	 * @author Piotrek
	 */
	public void RemoveSelectedItem()
	{
		Connection selectedConnection = (Connection)connectionsList.getSelectedValue();
		
		if (selectedConnection != null)
		{
			listModel.removeElement(selectedConnection);
			
			World.GetInstance().RemoveConnection(selectedConnection);
			
			worldView.repaint();
		}
		
		editButton.setEnabled(listModel.size() > 0);
		removeButton.setEnabled(listModel.size() > 0);
	}
	
	/**
	 * Loads the connections from the world.
	 * @author Piotrek
	 */
	public void LoadConnections()
	{	
		listModel.clear();
		
		Iterator connections = World.GetInstance().GetConnections();
		
		while (connections.hasNext())
		{
			Connection connection = (Connection)connections.next();
			
			listModel.addElement(connection);
		}
		
		editButton.setEnabled(listModel.size() > 0);
		removeButton.setEnabled(listModel.size() > 0);
	}
	
	
	/**
	 * @author Piotrek
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		// edit button
		if (e.getSource() == editButton)
		{
			EditSelectedConnection();
		}
		
		// remove button
		if (e.getSource() == removeButton)
		{
			RemoveSelectedItem();
		}
	}
	
	/**
	 * Edits the currently set connection.
	 * @author Piotrek
	 */
	private void EditSelectedConnection()
	{
		Connection connection = GetSelectedConnection();
		
		if (connection == null)
			return;
		
		ConnectionTimetableCreator timetableCreator = new ConnectionTimetableCreator("Edit a Timetable", connection.GetName(), connection.GetStations(), 0, connection);
		
		timetableCreator.setVisible(true);
		
		if (timetableCreator.GetDialogResult() == ConnectionTimetableCreator.OK)
		{
			// creator has added the connection
			// so we need to remove the old instance
			World.GetInstance().RemoveConnection(connection);
						
			connectionsList.clearSelection();
						
			int position = listModel.indexOf(connection);
			
			listModel.removeElement(connection);
			listModel.insertElementAt(timetableCreator.GetAddedConnection(), position);
			
			connectionsList.repaint();
			
			worldView.repaint();
		}
	}
	
	/**
	 * Gets the currently selected conection.
	 * @author Piotrek
	 * @return A currently selected connection or a null value if there is none.
	 */
	public Connection GetSelectedConnection()
	{
		Connection selectedConnection = (Connection)connectionsList.getSelectedValue();
		
		if (selectedConnection != null)
		{
			return selectedConnection;
		}
		
		return null;
	}
	
	/**
	 * Repaints the view, so the selected connection might be drawn.
	 * @author Piotrek
	 * @see javax.swing.event.ListSelectionListener#valueChanged(javax.swing.event.ListSelectionEvent)
	 */
	public void valueChanged(ListSelectionEvent e)
	{
		worldView.repaint();
	}
}
