/*
 * Created on 2005-04-09
 */
package gui.controls;

import gui.TravelSchedulerWorldEditor;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.ParseException;
import java.util.ArrayList;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import world.Connection;
import world.TravelTime;

/**
 * A panel used to define the times of arrivals and departures for every station of a connection.
 * @author Piotrek
 */
public final class StationsTimetableEditor extends JPanel implements ActionListener
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param stations An ordered list of all the stations in the connection (at least two). Ignored if connection is not null.
	 * @param meanOfTransportation A mean of transportation. This is required only to count the default times of travel. Ignored if connection is not null.
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @param connection A connection to be load for the edition. Set it to null if creating a new connection.
	 */
	public StationsTimetableEditor(ArrayList stations, int meanOfTransportation, Connection connection)
	{
		super(null);
		
		if (connection != null)
		{
			stations = connection.GetStations();
			
			meanOfTransportation = connection.GetMeanOfTransportation();
			
			this.connection = connection;
		}
		
		if (stations == null)
			throw new NullPointerException("Stations list must not be null in StationsTimetableEditor.StationsTimetableEditor().");
		
		if (stations.size() < 2)
			throw new IllegalArgumentException("Stations list must contain at least two stations in StationsTimetableEditor.StationsTimetableEditor().");
				
		setBackground(Color.BLUE);
		
		this.stations = stations;
		
		this.meanOfTransportation = meanOfTransportation;
		
		AddHeader();
		
		AddList();
		
		AddConvertionButtons();
		
		setVisible(true);
	}
	
	/**
	 * A connection to be load for the edition. Null if creating a new connection.
	 * @author Piotrek
	 */
	private Connection connection = null;
	
	/**
	 * A list of all the stations the times are to be set for.
	 * @author Piotrek
	 */
	private ArrayList stations = null;
	
	/**
	 * Adds a header over the list with the names of columns.
	 * @author Piotrek
	 */
	private void AddHeader()
	{
		// station number
		JLabel label = new JLabel("Number", SwingConstants.CENTER);
		
		label.setBackground(Color.ORANGE);
		label.setBounds(20, 20, 80, 20);
		label.setOpaque(true);
		
		add(label);
		
		// station name
		label = new JLabel("Station Name", SwingConstants.CENTER);
		
		label.setBackground(Color.LIGHT_GRAY);
		label.setBounds(100, 20, 250, 20);
		label.setOpaque(true);
		
		add(label);
		
		// arrival
		label = new JLabel("Arrival Time", SwingConstants.CENTER);
		
		label.setBackground(Color.ORANGE);
		label.setBounds(350, 20, 200, 20);
		label.setOpaque(true);
		
		add(label);
		
		// departure
		label = new JLabel("Departure Time", SwingConstants.CENTER);
		
		label.setBackground(Color.LIGHT_GRAY);
		label.setBounds(550, 20, 200, 20);
		label.setOpaque(true);
		
		add(label);
	}
	
	/**
	 * A mean of transportation the stations are for.
	 * @author Piotrek
	 */
	private int meanOfTransportation = -1;
	
	/**
	 * Adds a list for editing times of arrivals and departures for all the stations.
	 * @author Piotrek
	 */
	private void AddList()
	{
		list = new StationsTimetableListView(stations, meanOfTransportation, connection);
		
		listPane = new JScrollPane(list);
		
		add(listPane);
		
		listPane.setBounds(20, 40, 730, 430);
	}
		
	/**
	 * A list view of stations.
	 * @author Piotrek
	 */
	private StationsTimetableListView list = null;
	
	/**
	 * The scroll pane making the list scrollable.
	 * @author Piotrek
	 */
	private JScrollPane listPane;
	
	/**
	 * Gets a time in minutes since midnight of departure from the station with the given index.
	 * @author Piotrek
	 * @param index An index of the station.
	 * @return A time in minutes since midnight of departure from the station with the given index or -1 if there was any parse error.
	 */
	public int GetDepartureTime(int index)
	{
		if (index < 0 || index >= stations.size())
			throw new IllegalArgumentException("The index in the StationsTimetableEditor.GetDepartureTime() is out of range.");
		
		ArrayList arrivalTextFields = list.GetArrivalTextFields();
		ArrayList departureTextFields = list.GetDepartureTextFields();
		
		int time = 0;
		
		int startTime = 0;
		
		try
		{
			for (int i = 0; i <= index; ++i)
			{
				String arrivalTimeString = ((JTextField)arrivalTextFields.get(i)).getText();
				String departureTimeString = ((JTextField)departureTextFields.get(i)).getText();
				
				if (i == 0)
					startTime = TravelTime.ParseTime(arrivalTimeString);
				
				if (arrivalTimeString.startsWith("+"))
					time = TravelTime.ParseDuration(arrivalTimeString) + startTime;
				else
					time = TravelTime.ParseTime(arrivalTimeString);
					
				if (departureTimeString.startsWith("+"))
					time = TravelTime.ParseDuration(departureTimeString) + startTime;
				else
					time = TravelTime.ParseTime(departureTimeString);
			}
		}
		
		catch (ParseException xcp)
		{
			return -1;
		}
		
		return time;
	}
	
	/**
	 * Gets a time in minutes since midnight of arrival to the station with the given index.
	 * @author Piotrek
	 * @param index An index of the station.
	 * @return A time in minutes since midnight of arriavl to the station with the given index or -1 if there was any parse error.
	 */
	public int GetArrivalTime(int index)
	{
		if (index < 0 || index >= stations.size())
			throw new IllegalArgumentException("The index in the StationsTimetableEditor.GetArrivalTime() is out of range.");
		
		ArrayList arrivalTextFields = list.GetArrivalTextFields();
		ArrayList departureTextFields = list.GetDepartureTextFields();
		
		int time = 0;
		
		int startTime = 0;
		
		try
		{
			for (int i = 0; i <= index; ++i)
			{
				String arrivalTimeString = ((JTextField)arrivalTextFields.get(i)).getText();
				String departureTimeString = ((JTextField)departureTextFields.get(i)).getText();
				
				if (i == 0)
					startTime = TravelTime.ParseTime(arrivalTimeString);
				
				if (arrivalTimeString.startsWith("+"))
					time = TravelTime.ParseDuration(arrivalTimeString) + startTime;
				else
					time = TravelTime.ParseTime(arrivalTimeString);
					
				if (i != index)
				{
					if (departureTimeString.startsWith("+"))
						time = TravelTime.ParseDuration(departureTimeString) + startTime;
					else
						time = TravelTime.ParseTime(departureTimeString);
				}
			}
		}
		
		catch (ParseException xcp)
		{
			return -1;
		}
		
		return time;
	}
	
	/**
	 * Adds the buttons for applying the convertions between `+hh:mm' and `hh:mm' forms.
	 * @author Piotrek
	 * @author Piotrek
	 *
	 */
	private void AddConvertionButtons()
	{
		// shift form
		convertToShiftTimeForm = new JButton("Convert to shift form");
		
		convertToShiftTimeForm.setBounds(20, 480, 200, 25);
		convertToShiftTimeForm.setBackground(Color.ORANGE);
		convertToShiftTimeForm.addActionListener(this);
		
		add(convertToShiftTimeForm);
		
		// absolute form
		convertToAbsoluteTimeForm = new JButton("Convert to absolute form");
		
		convertToAbsoluteTimeForm.setBounds(240, 480, 200, 25);
		convertToAbsoluteTimeForm.setBackground(Color.ORANGE);
		convertToAbsoluteTimeForm.addActionListener(this);
		
		add(convertToAbsoluteTimeForm);
	}
	
	/**
	 * A buttong used to convert the timetable to the `+hh:mm' form.
	 * @author Piotrek
	 */
	private JButton convertToShiftTimeForm = null;
	
	/**
	 * A buttong used to convert the timetable to the `hh:mm' form.
	 * @author Piotrek
	 */
	private JButton convertToAbsoluteTimeForm = null;
		
	/**
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 * @author Piotrek
	 */
	public void actionPerformed(ActionEvent e)
	{
		if (e.getSource() == convertToShiftTimeForm)
		{
			OnConvertToShiftTimeForm();
			
			return;
		}
		
		if (e.getSource() == convertToAbsoluteTimeForm)
		{
			OnConvertToAbsoluteTimeForm();
			
			return;
		}
	}
	
	/**
	 * A handler for the convert to absolute time form button.
	 * It converts the timetable to the absolute time format.
	 * @author Piotrek
	 * @see #actionPerformed(ActionEvent)
	 */
	private void OnConvertToAbsoluteTimeForm()
	{
		if (!Validate())
			return;
		
		list.SetDepartureTime(TravelTime.GetTime(GetDepartureTime(0)), 0);
		
		for (int i = 1; i < stations.size(); ++i)
		{
			list.SetArrivalTime(TravelTime.GetTime(GetArrivalTime(i)), i);
			list.SetDepartureTime(TravelTime.GetTime(GetDepartureTime(i)), i);
		}
	}
	
	/**
	 * A handler for the convert to shift time form button. 
	 * It converts times to the shift form.
	 * @author Piotrek
	 * @see #actionPerformed(ActionEvent)
	 */
	private void OnConvertToShiftTimeForm()
	{
		if (!Validate())
			return;
		
		int startTime = GetArrivalTime(0);
		
		list.SetDepartureTime(TravelTime.GetShiftTime(GetDepartureTime(0) - startTime), 0);
		
		for (int i = 1; i < stations.size(); ++i)
		{
			list.SetArrivalTime(TravelTime.GetShiftTime(GetArrivalTime(i) - startTime), i);
			list.SetDepartureTime(TravelTime.GetShiftTime(GetDepartureTime(i) - startTime), i);
		}
	}
	
	/**
	 * Validates the timetable.
	 * @author Piotrek
	 * @param errorsList A list of errors (strings). If there are any, they will be added to this list.
	 * @return A bool value indicating whether the timetable is OK.
	 */
	public boolean ValidateTimetable(ArrayList errorsList)
	{
		if (errorsList == null)
			throw new NullPointerException("Errors list must not be null in StationsTimetableEditor.ValidateTimetable().");
		
		boolean fOK = true;
		
		int previousDepartureTime = -1;
		
		for (int i = 0; i < stations.size(); ++i)
		{
			int arrivalTime = GetArrivalTime(i);
			int departureTime = GetDepartureTime(i);
			
			if (arrivalTime == -1)
			{
				fOK = false;
				
				errorsList.add("The arrival time for the #" + (i + 1) + " station is incorrect. All next times might be erroneous, because of that.");
			}
			
			if (departureTime == -1)
			{
				fOK = false;
				
				errorsList.add("The departure time for the #" + (i + 1) + " station is incorrect. All next times might be erronous, because of that.");
			}
			
			if (previousDepartureTime == -1)
				previousDepartureTime = departureTime;
			else
			{
				if (arrivalTime != -1 && previousDepartureTime != -1)
					if (arrivalTime <= previousDepartureTime)
					{
						fOK = false;
						
						errorsList.add("The arrival time for the #" + (i + 1) + " station is not after the departure from the previous station.");
					}
			}
			
			if (arrivalTime != -1 && departureTime != -1)
				if (arrivalTime > departureTime)
				{
					fOK = false;
					
					errorsList.add("The arrival time for the #" + (i + 1) + " station is after the departure time for that station.");
				}
		}
		
		return fOK;
	}
	
	/**
	 * Validates the timetable. It shows the errors in the message box.
	 * @author Piotrek
	 * @return A bool value indicating whether the timetable is OK.
	 */
	private boolean Validate()
	{
		ArrayList errorsList = new ArrayList();
		
		if (!ValidateTimetable(errorsList))
		{
			TravelSchedulerWorldEditor.ShowErrorsList("The timetable is erroneous:", errorsList);
						
			return false;
		}		
		
		return true;
	}
}
