/*
 * Created on 2005-05-23
 */
package scheduler;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import world.Connection;
import world.Location;
import world.MeanOfTransportation;
import world.Station;
import world.TravelTime;

/**
 * A class used to store a single task in the tasks list
 * @author Piotrek
 */
public final class ConnectionsFinderTask implements Cloneable
{	
	/**
	 * Constructor
	 * @author Piotrek
	 */
	public ConnectionsFinderTask()
	{
		// does nothing
	}
	
	/**
	 * A list of visited stations used to detect the cycles.
	 * @author Piotrek
	 */
	private ArrayList visitedStations = new ArrayList();
	
	/**
	 * Adds a station to the list of visited stations.
	 * @author Piotrek
	 * @param station A station to be added.
	 */
	public void AddVisitedStation(Station station)
	{
		if (station == null)
			throw new NullPointerException("Station must not be null in the ConnectionsFinderTask.AddVisitedStation().");
		
		visitedStations.add(station);
	}
	
	/**
	 * Checks whether the given station has been visited before.
	 * @author Piotrek
	 * @param station A station to be tested.
	 * @return A boolean value indicating whether the given station has been visited before.
	 */
	public boolean HasVisitedStation(Station station)
	{
		if (station == null)
			throw new NullPointerException("Station must not be null in ConnectionsFinderTask.HasVisitedStation().");
		
		return visitedStations.contains(station);
	}
	
	/**
	 * A list of currently so far added connection chunks within the current travel part.
	 * @author Piotrek
	 */
	private ArrayList currentConnectionChunks = new ArrayList();
	
	/**
	 * A list of idecies of the chunks with at least one of the start and end dates being unset. It holds indecies as the Integers.
	 * @author Piotrek
	 */
	private ArrayList unsetDatesChunksIndecies = new ArrayList();
	
	/**
	 * Gets an iterator to the list of chunks with the unset dates. The indecies are in the decreasing order.
	 * @author Piotrek
	 * @return An iterator to the list of chunks with the unset dates. The indecies are in the decreasing order.
	 */
	public Iterator GetUnsetDatesChunksIndecies()
	{
		return unsetDatesChunksIndecies.iterator();
	}
	
	/**
	 * Gets a boolean value indicating whether this task has some connection chunks with unset dates.
	 * @author Piotrek
	 * @return A boolean value indicating whether this task has some connection chunks with unset dates.
	 */
	public boolean HasUnsetDatesChunks()
	{
		return unsetDatesChunksIndecies.size() > 0;
	}
	
	/**
	 * Gets the connection chunk with the given index.
	 * @author Piotrek
	 * @param chunkIndex An index of the connection chunk to be returned.
	 * @return A connection chunk with the given index.
	 */
	public TravelPartConnectionChunk GetChunk(int chunkIndex)
	{
		if (chunkIndex < 0 || chunkIndex >= currentConnectionChunks.size())
			throw new IndexOutOfBoundsException("Index out of range in the ConnectionsFinderTask.GetChunk().");
		
		return (TravelPartConnectionChunk)currentConnectionChunks.get(chunkIndex);
	}
	
	/**
	 * Adds a new connection chunk at the end of the current list of connection chunks. If it has at least one of the start and the end dates unset, its index is added to the special list of indecies of chunks with unset dates.
	 * @author Piotrek
	 * @param connectionChunk A connection chunk to be added.
	 */
	public void AddConnectionChunk(TravelPartConnectionChunk connectionChunk)
	{
		// add test
		if (connectionChunk == null)
			throw new NullPointerException("Connection chunk must not be null in the ConenctionsFinderTask.AddConnectionChunk().");
						
		int chunksCount = currentConnectionChunks.size();
		
		// checking if the new one starts where the old ones end
		if (chunksCount > 0)
		{
			TravelPartConnectionChunk lastChunk = (TravelPartConnectionChunk)currentConnectionChunks.get(chunksCount - 1);
			
			if (!lastChunk.GetDestination().IsEqual(connectionChunk.GetOrigin()))
				throw new IllegalArgumentException("The new connection chunk must end where the last of the previous ones ends in the ConenctionsFinderTask.AddConnectionChunk().");
		}
		
		// checking if it means a change
		if (connectionChunk.GetMeanOfTransportation() != MeanOfTransportation.WALK && chunksCount > 0)
		{
			TravelPartConnectionChunk oldChunks = null;
			
			int i = chunksCount - 1;
			
			while (i >= 0 && (oldChunks = (TravelPartConnectionChunk)currentConnectionChunks.get(i)).GetMeanOfTransportation() == MeanOfTransportation.WALK)
				--i;
			
			if (oldChunks.GetMeanOfTransportation() != MeanOfTransportation.WALK)
				++changesCount;
		}
		
		currentConnectionChunks.add(connectionChunk);
		
		// checking if it is a new mean of transportation
		Integer box = new Integer(connectionChunk.GetMeanOfTransportation());

		if (!usedMeansOfTransportation.contains(box))
			usedMeansOfTransportation.add(box);
		
		// checking if it has an unset date		
		if (connectionChunk.GetStartDate() == null || connectionChunk.GetEndDate() == null)
			unsetDatesChunksIndecies.add(0, new Integer(chunksCount));
	}
	
	/**
	 * Gets a current list of connection chunks.
	 * @author Piotrek
	 * @return An iterator for the current list of connection chunks.
	 */
	public Iterator GetConnectionChunks()
	{
		return currentConnectionChunks.iterator();
	}
	
	/**
	 * Gets the first connection chunk in the list.
	 * @author Piotrek
	 * @return The first connection chunk in the list or null value if there is none.
	 */
	public TravelPartConnectionChunk GetFirstConnectionChunk()
	{
		if (currentConnectionChunks.size() > 0)
			return (TravelPartConnectionChunk)currentConnectionChunks.get(0);
		
		return null;
	}
	
	/**
	 * Gets the last connection chunk in the list.
	 * @author Piotrek
	 * @return The last connection chunk in the list or null value if there is none.
	 */
	public TravelPartConnectionChunk GetLastConnectionChunk()
	{
		int listSize = currentConnectionChunks.size();
		
		if (listSize > 0)
			return (TravelPartConnectionChunk)currentConnectionChunks.get(listSize - 1);
		
		return null;
	}
	
	/**
	 * Gets the last connection chunk in the list with the end date set.
	 * @author Piotrek
	 * @return The last connection chunk in the list with the end date set or null value if there is none.
	 */
	public TravelPartConnectionChunk GetLastConnectionChunkWithEndDate()
	{
		int i = currentConnectionChunks.size() - 1;
		
		while (i >= 0 && ((TravelPartConnectionChunk)currentConnectionChunks.get(i)).GetEndDate() == null)
			--i;
		
		if (i >= 0 && ((TravelPartConnectionChunk)currentConnectionChunks.get(i)).GetEndDate() != null)
			return (TravelPartConnectionChunk)currentConnectionChunks.get(i);
		
		return null;
	}
	
	/**
	 * A current real connection the work is done upon.
	 * @author Piotrek
	 */
	public Connection connection = null;
		
	/**
	 * A current work station (within the current connection).
	 * @author Piotrek
	 */
	public Station currentStation = null;
	
	/**
	 * A starting station used to remember the station where the new connection chunk should start.
	 * @author Piotrek
	 */
	public Station startStation = null;
	
	/**
	 * A day of week the current connection starts on.
	 * @author Piotrek
	 * @see Connection#MONDAY
	 * @see Connection#TUESDAY
	 * @see Connection#WEDNESDAY
	 * @see Connection#THURSDAY
	 * @see Connection#FRIDAY
	 * @see Connection#SATURDAY
	 * @see Connection#SUNDAY
	 */
	public char dayOfWeek = 0;
	
	/**
	 * A date of departure from the start station. Null value means there is no date availble at the moment.
	 * @author Piotrek
	 */
	public Date startStationDepartureDate = null;
	
	/**
	 * A clone has the reference to the same connection chunks except for the chunks with at least one unset date (those are new copies).
	 * @see java.lang.Object#clone()
	 * @author Piotrek
	 */
	public Object clone()
	{
		ConnectionsFinderTask copy = new ConnectionsFinderTask();
		
		// connection chunks
		
		Iterator iterator = currentConnectionChunks.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			if (chunk.GetStartDate() != null && chunk.GetEndDate() != null)
				copy.AddConnectionChunk(chunk);
			else
				copy.AddConnectionChunk((TravelPartConnectionChunk)chunk.clone());
		}
		
		// visited stations
		iterator = visitedStations.iterator();
		
		while (iterator.hasNext())
		{
			Station station = (Station)iterator.next();
			
			copy.AddVisitedStation(station);
		}
		
		// other things
		copy.connection = connection;
		
		copy.fDirect = fDirect;
		
		copy.startStation = startStation;
		copy.currentStation = currentStation;
		
		copy.dayOfWeek = dayOfWeek;
		
		copy.startStationDepartureDate = startStationDepartureDate;
		
		return copy;
	}
	
	/**
	 * A number of changes in the list of the connection chunks.
	 * @author Piotrek
	 */
	private int changesCount = 0;
	
	/**
	 * Gets a number of changes in the list of the connection chunks.
	 * @author Piotrek
	 * @return A number of changes in the list of the connection chunks.
	 */
	public int GetChangesCount()
	{
		return changesCount;
	}
	
	/**
	 * Clears the list of visited stations.
	 * @author Piotrek
	 */
	public void ClearVisitedStations()
	{
		visitedStations.clear();
	}
	
	/**
	 * Sets the current connection, day of week, current station and the start station to the same values as of the other one's.
	 * @author Piotrek
	 * @param otherTask A source of settings.
	 */
	public void SetCurrentWorkingData(ConnectionsFinderTask otherTask)
	{
		if (otherTask == null)
			throw new NullPointerException("Other task must not be null in the ConnectionsFinderTask.SetCurrentWorkingData().");
				
		connection = otherTask.connection;
		dayOfWeek = otherTask.dayOfWeek;
		
		startStation = otherTask.startStation;
		currentStation = otherTask.currentStation;
		
		startStationDepartureDate = otherTask.startStationDepartureDate;
		
		fDirect = otherTask.fDirect;
	}
	
	/**
	 * Gets a number of connection chunks contained by this task at the moment.
	 * @author Piotrek
	 * @return A number of connection chunks contained by this task at the moment.
	 */
	public int GetConnectionChunksCount()
	{
		return currentConnectionChunks.size();
	}
	
	/**
	 * Indicates whether the connection is a direct connection to the destination station.
	 * @author Piotrek
	 */
	public boolean fDirect = false;
	
	/**
	 * Gets a total duration of the travel stored in this task.
	 * @author Piotrek
	 * @return Duration in minutes or -1 if not all dates are set.
	 */
	public int GetTotalDuration()
	{
		int duration = 0;
		
		Iterator iterator = currentConnectionChunks.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			Date startDate = chunk.GetStartDate();
			Date endDate = chunk.GetEndDate();
			
			if (startDate == null || endDate == null)
				return -1;
			
			int startTime = (int)(startDate.getTime() / 60000);
			int endTime = (int)(endDate.getTime() / 60000);
			
			duration += endTime - startTime;
		}
		
		return duration;
	}

	/**
	 * Checks whether this task is similar to the given one. Tasks are similar if they have the same connection chunks. Only dates may differ (except for the start and the end dates).
	 * All chunks must have all the dates set. This method does not check if it is so.
	 * @author Piotrek
	 * @param otherTask The other task this one is to be compared.
	 * @return A value indicating whether this task is similar to the given one.
	 */
	public boolean IsSimilar(ConnectionsFinderTask otherTask)
	{
		if (otherTask == null)
			throw new NullPointerException("Other task must not be null in ConnectionsFinderTask.IsSimilar().");
		
		// simple tests
		if (currentConnectionChunks.size() != otherTask.GetConnectionChunksCount())
			return false;
		
		if (!GetFirstConnectionChunk().GetStartDate().equals(otherTask.GetFirstConnectionChunk().GetStartDate()))
			return false;
		
		if (!GetLastConnectionChunk().GetEndDate().equals(otherTask.GetLastConnectionChunk().GetEndDate()))
			return false;
		
		// comparing chunks
		Iterator myIterator = currentConnectionChunks.iterator();
		Iterator otherIterator = otherTask.GetConnectionChunks();
		
		while (myIterator.hasNext()) // both tasks have the same number of chunks
		{
			TravelPartConnectionChunk myChunk = (TravelPartConnectionChunk)myIterator.next();
			TravelPartConnectionChunk otherChunk = (TravelPartConnectionChunk)otherIterator.next();
			
			if (myChunk.GetMeanOfTransportation() != otherChunk.GetMeanOfTransportation())
				return false;
			
			if (!myChunk.GetOrigin().IsEqual(otherChunk.GetOrigin()))
				return false;
			
			if (!myChunk.GetDestination().IsEqual(otherChunk.GetDestination()))
				return false;
		}
		
		return true;
	}
	
	/**
	 * Gets a duration of the longest change.
	 * @author Piotrek
	 * @return A duration of the longest change in minutes or -1 if there is no change and/or if there is any unset date.
	 */
	public int GetMaxChangeDuration()
	{
		if (changesCount == 0)
			return -1;
		
		int maxChangeDuration = -1;
				
		TravelPartConnectionChunk previousChunk = null;
		TravelPartConnectionChunk beforePreviousChunk = null;
		
		Iterator iterator = currentConnectionChunks.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			Date startDate = chunk.GetStartDate();
			Date endDate = chunk.GetEndDate();
			
			if (startDate == null || endDate == null)
				return -1;
			
			if (previousChunk != null)
			{
				if (chunk.GetMeanOfTransportation() != MeanOfTransportation.WALK)
				{
					// change without the walk
					if (previousChunk.GetMeanOfTransportation() != MeanOfTransportation.WALK)
					{
						int changeDuration = TravelTime.GetDatesDifference(previousChunk.GetEndDate(), chunk.GetStartDate());
						
						if (changeDuration > maxChangeDuration)
							maxChangeDuration = changeDuration;
					}
					else
					// change with the walk
					if (beforePreviousChunk != null)
					{
						int changeDuration = TravelTime.GetDatesDifference(beforePreviousChunk.GetEndDate(), chunk.GetStartDate());
						
						if (changeDuration > maxChangeDuration)
							maxChangeDuration = changeDuration;
					}
				}
			}
			
			// step
			beforePreviousChunk = previousChunk;
			
			previousChunk = chunk;
		}
		
		return maxChangeDuration; 
	}
	
	/**
	 * Gets a duration of the average change.
	 * @author Piotrek
	 * @return A duration of the average change in minutes or -1 if there is no change and/or if there is any unset date.
	 */
	public int GetAverageChangeDuration()
	{
		if (changesCount == 0)
			return -1;
		
		int changesDurationSum = 0;
				
		TravelPartConnectionChunk previousChunk = null;
		TravelPartConnectionChunk beforePreviousChunk = null;
		
		Iterator iterator = currentConnectionChunks.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			Date startDate = chunk.GetStartDate();
			Date endDate = chunk.GetEndDate();
			
			if (startDate == null || endDate == null)
				return -1;
			
			if (previousChunk != null)
			{
				if (chunk.GetMeanOfTransportation() != MeanOfTransportation.WALK)
				{
					// change without the walk
					if (previousChunk.GetMeanOfTransportation() != MeanOfTransportation.WALK)
					{
						int changeDuration = TravelTime.GetDatesDifference(previousChunk.GetEndDate(), chunk.GetStartDate());
						
						changesDurationSum += changeDuration;
					}
					else
					// change with the walk
					if (beforePreviousChunk != null)
					{
						int changeDuration = TravelTime.GetDatesDifference(beforePreviousChunk.GetEndDate(), chunk.GetStartDate());
						
						changesDurationSum += changeDuration;
					}
				}
			}
			
			// step
			beforePreviousChunk = previousChunk;
			
			previousChunk = chunk;
		}
				
		return changesDurationSum / changesCount; 
	}
	
	/**
	 * Gets the total distance of the travel described by this task.
	 * @author Piotrek
	 * @return A total distance of the travel.
	 */
	public double GetTotalDistance()
	{
		double distance = 0;
		
		Iterator iterator = currentConnectionChunks.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			distance += Location.GetDistance(chunk.GetOrigin(), chunk.GetDestination());
		}
		
		return distance;
	}
	
	/**
	 * A current list of used means of transportation.
	 * @author Piotrek
	 */
	private ArrayList usedMeansOfTransportation = new ArrayList(6);
	
	/**
	 * Gets a list of used means of transportation.
	 * @author Piotrek
	 * @return A list of used means of transportation. Each element on the list is an Integer bocing the symbol of a mean of transportation.
	 * @see MeanOfTransportation#AIRPLANE
	 * @see MeanOfTransportation#BUS
	 * @see MeanOfTransportation#SHIP
	 * @see MeanOfTransportation#TAXI
	 * @see MeanOfTransportation#TRAIN
	 * @see MeanOfTransportation#WALK
	 */
	public ArrayList GetUsedMeansOfTransportation()
	{
		return (ArrayList)usedMeansOfTransportation.clone();
	}
	
	/**
	 * An extra time in minutes the traveler is at the station before the departure (after the change).
	 * For example, this might be a time one need to be at the airport before the flight - assuming one got to the airport by bus, 
	 * there is a change time after getting out of the bus for a walk to the airport and just then there is this extra time.
	 * @author Piotrek
	 */
	public int extraTimeBeforeDeparture = 0;
	
	/**
	 * This parameter might be set by the judgement module basing on the history.
	 * It is provided only for the convinience.
	 * @author Piotrek
	 */
	public double historyBasedJudgement = 0;
}
