/*
 * Created on 2005-05-22
 */
package scheduler;

import java.util.Date;
import world.Connection;
import world.Location;

/**
 * A class describing a piece of a real connection used to fill a travel part.
 * A list of such chunks creates a final travel plan.
 * @author Piotrek
 */
public final class TravelPartConnectionChunk extends TravelPart implements Cloneable
{	
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param meanOfTransportation A mean of transportation.
	 * @param origin An origin location of this part of travel.
	 * @param destination A destination location of this part of travel.
	 */
	public TravelPartConnectionChunk(int meanOfTransportation, Location origin, Location destination)
	{
		super(meanOfTransportation, origin, destination);
	}
	
	/**
	 * The start date of this travel part. Null value means it must be computed basing on the subsequent connection chunk at the later time.
	 * @author Piotrek 
	 */
	private Date startDate = null;
	
	/**
	 * The end date of this travel part. Null value means it must be computed basing on the subsequent connection chunk at the later time.
	 * @author Piotrek 
	 */
	private Date endDate = null;
	
	/**
	 * A connection this travel part is taken by.
	 * @author Piotrek
	 */
	private Connection realConnection = null;
	
	/**
	 * An index of the origin station in the real connection's stations list. The value of -1 means this part of travel is taken by a taxi or by a walk.
	 * @author Piotrek
	 */
	private int originStationIndex = -1;
	
	/**
	 * An index of the destination station in the real connection's stations list. The value of -1 means this part of travel is taken by a taxi or by a walk.
	 * @author Piotrek
	 */
	private int destinationStationIndex = -1;

	/**
	 * Gets the index of the origin station in the real connection's stations list.
	 * @author Piotrek
	 * @return The index of the origin station in the real connection's stations list.  The value of -1 means this part of travel is taken by a taxi or by a walk.
	 */
	public int GetOriginStationIndex()
	{
		return originStationIndex;
	}
	
	/**
	 * Sets the index of the origin station in the real connection's stations list.
	 * @author Piotrek
	 * @param originStationIndex The index of the origin station in the real connection's stations list.
	 */
	public void SetOriginStationIndex(int originStationIndex)
	{
		if (realConnection == null)
			throw new RuntimeException("Set the connection firts in TravelPartConnectionChunk.SetOriginStationIndex().");
		
		if (originStationIndex < 0 || originStationIndex >= realConnection.GetStationsCount())
			throw new IndexOutOfBoundsException("Origin stations' index is out of range in TravelPartConnectionChunk.SetOriginStationIndex().");
		
		this.originStationIndex = originStationIndex;
	}
	
	/**
	 * Gets the index of the destination station in the real connection's stations list.
	 * @author Piotrek
	 * @return The index of the destination station in the real connection's stations list.  The value of -1 means this part of travel is taken by a taxi or by a walk.
	 */
	public int GetDestinationStationIndex()
	{
		return destinationStationIndex;
	}
	
	/**
	 * Sets the index of the destination station in the real connection's stations list.
	 * @author Piotrek
	 * @param destinationStationIndex The index of the destination station in the real connection's stations list.
	 */
	public void SetDestinationStationIndex(int destinationStationIndex)
	{
		if (realConnection == null)
			throw new RuntimeException("Set the connection firts in TravelPartConnectionChunk.SetDestinationStationIndex().");
		
		if (destinationStationIndex < 0 || destinationStationIndex >= realConnection.GetStationsCount())
			throw new IndexOutOfBoundsException("Destination stations' index is out of range in TravelPartConnectionChunk.SetDestinationStationIndex().");
		
		this.destinationStationIndex = destinationStationIndex;
	}
	
	/**
	 * Gets a reference to the real connection this travel part is taken by.
	 * @author Piotrek
	 * @return The reference to the real connection this travel part is taken by. A null value means there is none connection (in case of a walk or a taxi).
	 */
	public Connection GetRealConnection()
	{
		return realConnection;
	}
	
	/**
	 * Sets the reference to the real connection this travel part is taken by.
	 * @author Piotrek
	 * @param realConnection A reference to the real connection part this travel part is taken by. A null value means there is no connection (in case of a walk or a taxi).
	 */
	public void SetRealConnection(Connection realConnection)
	{
		this.realConnection = realConnection;
	}
	
	/**
	 * Gets the start date of this travel part. 
	 * @author Piotrek
	 * @return A start date of this travel part. Null value means it must be computed basing on the subsequent connection chunk at the later time.
	 */
	public Date GetStartDate()
	{
		return startDate;
	}
	
	/**
	 * Sets the start date of this travel part. 
	 * @author Piotrek
	 * @param startDate A start date of this travel part. Null value means it must be computed basing on the subsequent connection chunk at the later time.
	 */
	public void SetStartDate(Date startDate)
	{
		this.startDate = startDate;
	}
	
	/**
	 * Gets the end date of this travel part. 
	 * @author Piotrek
	 * @return An end date of this travel part. Null value means it must be computed basing on the subsequent connection chunk at the later time.
	 */
	public Date GetEndDate()
	{
		return endDate;
	}
	
	/**
	 * Sets the end date of this travel part. 
	 * @author Piotrek
	 * @param endDate An end date of this travel part. Null value means it must be computed basing on the subsequent connection chunk at the later time.
	 */
	public void SetEndDate(Date endDate)
	{
		this.endDate = endDate;
	}
	
	/**
	 * Gets a copy of this connection chunk.
	 * @author Piotrek
	 * @return A copy of this connection chunk.
	 */
	public Object clone()
	{
		TravelPartConnectionChunk copy = new TravelPartConnectionChunk(GetMeanOfTransportation(), GetOrigin(), GetDestination());
	
		copy.SetRealConnection(realConnection);
		
		copy.SetOriginStation(GetOriginStation());
		
		if (originStationIndex != -1)
			copy.SetOriginStationIndex(originStationIndex);
		
		copy.SetDestinationStation(GetDestinationStation());
		
		if (destinationStationIndex != -1)
			copy.SetDestinationStationIndex(destinationStationIndex);
			
		if (startDate != null)
			copy.SetStartDate((Date)startDate.clone());
		else
			copy.SetStartDate(null);
		
		if (endDate != null)
			copy.SetEndDate((Date)endDate.clone());
		else
			copy.SetEndDate(null);
		
		return copy;
	}
	
	/**
	 * Calls GetRevertedConnectionChunk().
	 * @see scheduler.TravelPart#GetRevertedPart()
	 * @see #GetRevertedConnectionChunk()
	 * @author Piotrek
	 */
	public TravelPart GetRevertedPart()
	{
		return GetRevertedConnectionChunk();
	}
	
	/**
	 * Gets a reverted connection chunk.
	 * @author Piotrek
	 * @return The reverted connection chunk.
	 */
	public TravelPartConnectionChunk GetRevertedConnectionChunk()
	{
		TravelPartConnectionChunk revertedOne = new TravelPartConnectionChunk(GetMeanOfTransportation(), GetDestination(), GetOrigin());
		
		revertedOne.SetRealConnection(realConnection);
		
		revertedOne.SetOriginStation(GetDestinationStation());
		
		if (realConnection != null)
			revertedOne.SetOriginStationIndex(destinationStationIndex);
		
		revertedOne.SetStartDate(endDate);
		
		revertedOne.SetDestinationStation(GetOriginStation());

		if (realConnection != null)
			revertedOne.SetDestinationStationIndex(originStationIndex);
		
		revertedOne.SetEndDate(startDate);
		
		return revertedOne;
	}
	
	/**
	 * An extra time in minutes the traveler is at the station before the departure (after the change).
	 * For example, this might be a time one need to be at the airport before the flight - assuming one got to the airport by bus, 
	 * there is a change time after getting out of the bus for a walk to the airport and just then there is this extra time.
	 * @author Piotrek
	 */
	public int extraTimeBeforeDeparture = 0;
}
