/*
 * Created on 2005-09-05
 */
package scheduler.modules;

import java.util.ArrayList;
import java.util.Iterator;
import jess.Fact;
import jess.JessException;
import jess.Rete;
import jess.Token;
import jess.ValueVector;
import scheduler.ConnectionsFinderTask;
import scheduler.TestOptions;
import scheduler.TravelConstraints;
import scheduler.TravelPart;
import scheduler.modulesinterfaces.ChangesJudger;
import world.Location;
import world.Station;

/**
 * A JESS based changes judger deciding whether to drop the changes or not.
 * @author Piotrek
 */
public final class ChangesJessJudger implements ChangesJudger
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public ChangesJessJudger()
	{
		try
		{
			// loading rules
			rete.executeCommand("(batch \"./Data/JESS changes rules.txt\")");
						
			// this will define the query for the decision
			rete.executeCommand("(defquery searchDecision" +
								"\"Finds decision facts\"" +
								"(change ?x))");
		}
		
		catch (JessException xcp)
		{
			// this will cause the algorithm to stop
			throw new RuntimeException(xcp.getMessage());
		}
	}
	
	/**
	 * The Java Expert Shell System used to support rules genereting the proper means of transportation.
	 * @author Piotrek
	 */
	private Rete rete = new Rete();
	
	/**
	 * @author Piotrek
	 * @see ChangesJudger#SetTravelConstraints(scheduler.TravelConstraints)
	 */
	public void SetTravelConstraints(TravelConstraints constraints)
	{
		if (constraints == null)
			throw new NullPointerException("Constraints must not be null in ChangesJudgerStub.SetTravelConstraints().");
		
		travelConstraints = constraints;
	}
	
	/**
	 * The current travel travelConstraints.
	 * @author Piotrek
	 */
	private TravelConstraints travelConstraints = null;
	
	/**
	 * @see scheduler.modulesinterfaces.ChangesJudger#ShouldDropChange(scheduler.ConnectionsFinderTask, scheduler.ConnectionsFinderTask, scheduler.TravelPart, boolean)
	 * @author Piotrek
	 */
	public boolean ShouldDropChange(ConnectionsFinderTask task, ConnectionsFinderTask changeTask, TravelPart travelPart, boolean fRevert)
	{
		if (task == null)
			throw new NullPointerException("Task must not be null in ChangesJessJudger.ShouldDropChange().");
		
		if (changeTask == null)
			throw new NullPointerException("Change task must not be null in ChangesJessJudger.ShouldDropChange().");
		
		if (travelPart == null)
			throw new NullPointerException("Travel part must not be null in ChangesJessJudger.ShouldDropChange().");
		
		if (travelConstraints == null)
			throw new RuntimeException("Set the travel constrains first in ChangesJessJudger.ShouldDropChange().");
		
		boolean fShouldDrop = false;
		
		if (!TestOptions.ShouldUseRulesDroppingChanges())
			return false;
		
		try
		{			
			// reseting the rete
			rete.reset();
			
			// load data into the rete
			
			// min change time
			rete.executeCommand("(assert (minChangeTime " + travelConstraints.minChangeTime + "))");

			// are both connections direct
			if (task.fDirect)
				rete.executeCommand("(assert (currentConnectionIsDirect TRUE))");
			else
				rete.executeCommand("(assert (currentConnectionIsDirect FALSE))");
			
			if (changeTask.fDirect)
				rete.executeCommand("(assert (changeConnectionIsDirect TRUE))");
			else
				rete.executeCommand("(assert (changeConnectionIsDirect FALSE))");
			
			int changeActualStationIndex = changeTask.connection.GetStationIndex(changeTask.currentStation);
			int currentActualStationIndex = task.connection.GetStationIndex(task.currentStation);
			
			// times of travel to the destination station 
			
			int changeDestinationStationIndex = changeTask.connection.GetStationIndex(travelPart.GetDestinationStation());
			int currentDestinationStationIndex = task.connection.GetStationIndex(travelPart.GetDestinationStation());
					
			if (changeDestinationStationIndex != -1 && currentDestinationStationIndex != -1)
			{
				int changeTravelTime = changeTask.connection.GetArrivalTime(changeDestinationStationIndex) - changeTask.connection.GetDepartureTime(changeActualStationIndex);
				int currentTravelTime = task.connection.GetArrivalTime(currentDestinationStationIndex) - task.connection.GetDepartureTime(currentActualStationIndex);
				
				if (fRevert) // reverting - it will be enough to remove the negative sign
				{
					changeTravelTime *= -1;
					currentTravelTime *= -1;
				}
				
				rete.executeCommand("(assert (currentTimeToDestination " + currentTravelTime + "))");
				rete.executeCommand("(assert (changeTimeToDestination " + changeTravelTime + "))");
			}
			
			// if there are the next stations
			if ((!fRevert && currentActualStationIndex + 1 < task.connection.GetStationsCount() && changeActualStationIndex + 1 < changeTask.connection.GetStationsCount())
				|| (fRevert && currentActualStationIndex - 1 >= 0 && changeActualStationIndex - 1 >= 0))
			{
				rete.executeCommand("(assert (bothConnectionsHasNextStation TRUE))");
				
				int nextStationDelta = 1;
				
				if (fRevert)
					nextStationDelta = -1;
				
				Station currentNextStation = (Station)task.connection.GetStations().get(currentActualStationIndex + nextStationDelta);
				Station changeNextStation = (Station)changeTask.connection.GetStations().get(changeActualStationIndex + nextStationDelta);
				
				// if it is the same station
				if (currentNextStation.IsEqual(changeNextStation))
					rete.executeCommand("(assert (sameNextStations TRUE))");
				else
					rete.executeCommand("(assert (sameNextStations FALSE))");
				
				int changeTravelTime = changeTask.connection.GetArrivalTime(changeActualStationIndex + nextStationDelta) - changeTask.connection.GetDepartureTime(changeActualStationIndex);
				int currentTravelTime = task.connection.GetArrivalTime(currentActualStationIndex + nextStationDelta) - task.connection.GetDepartureTime(currentActualStationIndex);
				
				if (fRevert) // reverting - it will be enough to remove the negative sign
				{
					changeTravelTime *= -1;
					currentTravelTime *= -1;
				}
				
				rete.executeCommand("(assert (currentTimeToNextStation " + currentTravelTime + "))");
				rete.executeCommand("(assert (changeTimeToNextStation " + changeTravelTime + "))");				
			}
			else
				rete.executeCommand("(assert (bothConnectionsHasNextStation FALSE))");
			
			// distances
			double currentDistance = Location.GetDistance(task.currentStation.GetLocation(), travelPart.GetDestination());
			
			rete.executeCommand("(assert (currentConnectionDistance " + currentDistance + "))");
			
			// minimal distance from the next stations in the change connection

			ArrayList changeStations = changeTask.connection.GetStations();
			
			int changeCurrentStationIndex = changeTask.connection.GetStationIndex(changeTask.currentStation);

			int delta = 1;
			
			if (fRevert)
				delta = -1;
			
			double minDistance = Double.MAX_VALUE;
			
			int index = changeCurrentStationIndex + delta; // the next station
			
			boolean fWasAny = false;
											
			while (index >= 0 && index < changeStations.size())
			{
				double changeStationDistance = Location.GetDistance(((Station)changeStations.get(index)).GetLocation(), travelPart.GetDestination());
					
				if (changeStationDistance < minDistance)
					minDistance = changeStationDistance;
				
				fWasAny = true;
								
				index += delta;
			}
			
			if (fWasAny)
				rete.executeCommand("(assert (changeMinDistance " + minDistance + "))");
								
			// running the rules
			rete.executeCommand("(run)");
			
			// issue query for each mean of transportation
			
			// Run the query, store the result
		    rete.store("RESULT", rete.runQuery("searchDecision",
		                        new ValueVector()));
		    
		    rete.executeCommand("(store RESULT (run-query searchDecision))");

		    // Fetch the result (an Iterator).
		    Iterator resultsIterator = (Iterator) rete.fetch("RESULT").externalAddressValue(null);
		    
		    ArrayList changeDecisions = new ArrayList();

		    // Pick each element of the Iterator apart and store the
		    // interesting part in the ArrayListproperMeansOfTransportationv.
		    while (resultsIterator.hasNext())
		    {
				Token token = (Token)resultsIterator.next();

		    	// We want the second fact in the token - the first is the query trigger
		    	Fact fact = token.fact(1);

		    	// The first and only slot of this fact is the __data multislot.
		    	ValueVector multiSlot = fact.get(0).listValue(null);

		    	// The second element of this slot is the datum we're interested in.
		    	changeDecisions.add(multiSlot.get(0).stringValue(null));
		    }
		    
		    boolean fAccept = false;
		    boolean fReject = false;
		    
		    for (Iterator answers = changeDecisions.iterator(); answers.hasNext();)
		    {
		    	String changeDecision = (String)answers.next();
		    	
		    	if (changeDecision.compareTo("accept") == 0)
		    		fAccept = true;
		    	else
		    	if (changeDecision.compareTo("reject") == 0)
			    	fReject = true;
		       	else
		    		throw new RuntimeException("Unexpected decision received from the JESS system.");
		    }
		    
		    fShouldDrop = fReject && !fAccept;
		}
		
		catch (JessException xcp)
		{
			// this will stop the planning algorithm
			throw new RuntimeException("Line: " + xcp.getLineNumber() + "\nContext: " + xcp.getContext() + "\nData: " + xcp.getData() + "\nProgram text: " + xcp.getProgramText() + "\nMessage: " + xcp.getMessage());
		}
				
		return fShouldDrop;
	}
}
