/*
 * Created on 2005-04-11
 */
package world;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;

/**
 * A connection.
 * @author Piotrek
 */
public class Connection
{
	/**
	 * Constructor. It does not check if the times are in line with each other.
	 * @author Piotrek
	 * @param name A name for the connection.
	 * @param meanOfTransportation A mean of transportation for this connection.
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @param stations An ordered list of stations.
	 * @param arrivals An ordered list of arrival times (Integers representing the time.)
	 * @param departures An ordered list of departure times (Integers representing the time.)
	 * @param dayOfWeekMask A day of week mask.
	 * @see #MONDAY
	 * @see #TUESDAY
	 * @see #WEDNESDAY
	 * @see #THURSDAY
	 * @see #FRIDAY
	 * @see #SATURDAY
	 * @see #SUNDAY
	 */
	public Connection(String name, int meanOfTransportation, ArrayList stations, ArrayList arrivals, ArrayList departures, char dayOfWeekMask)
	{
		if (name == null)
			throw new NullPointerException("Name must not be null in Connection.Connection().");
		
		if (name.compareTo("") == 0)
			throw new IllegalArgumentException("Name must not be empty in Connection.Connection().");
		
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("The given mean of transportation is not a valid one in Connection.Conection().");
		
		if (stations.size() != arrivals.size() || stations.size() != departures.size())
			throw new IllegalArgumentException("The lists of stations, arrivals and departures have different sizes in Connection.Conection().");
		
		if (arrivals.get(0).getClass() != Integer.class)
			throw new IllegalArgumentException("A list of arrivals contain an object of unexpected type in Connection.Conection().");
		
		if (departures.get(0).getClass() != Integer.class)
			throw new IllegalArgumentException("A list of departures contain an object of unexpected type in Connection.Conection().");
		
		if (dayOfWeekMask == 0 || dayOfWeekMask >= 128)
			throw new IllegalArgumentException("A day of week mask has an incorrect value in Connection.Connection().");
		
		this.name = name;
		
		this.meanOfTransportation = meanOfTransportation;
		
		this.stations = stations;
		
		this.arrivals = arrivals;
		this.departures = departures;
		
		this.dayOfWeekMask = dayOfWeekMask;
	}
	
	/**
	 * A name of this connection.
	 * @author Piotrek
	 */
	private String name = null;
	
	/**
	 * Gets a name of this connection.
	 * @author Piotrek
	 * @return A name of this connection.
	 */
	public String GetName()
	{
		return name;
	}
	
	/**
	 * A Monday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char MONDAY = 1;
	
	/**
	 * A Tuesday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char TUESDAY = 2;
	
	/**
	 * A Wednesday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char WEDNESDAY = 4;
	
	/**
	 * A Thursday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char THURSDAY = 8;
	
	/**
	 * A Friday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char FRIDAY = 16;
	
	/**
	 * A Saturday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char SATURDAY = 32;
	
	/**
	 * A Sunday flag for the day of the week.
	 * @author Piotrek
	 */
	public static final char SUNDAY = 64;
	
	/**
	 * A day of week mask.
	 * @author Piotrek
	 */
	private char dayOfWeekMask;
	
	/**
	 * Gets the day of week mask.
	 * @author Piotrek
	 * @see #MONDAY
	 * @see #TUESDAY
	 * @see #WEDNESDAY
	 * @see #THURSDAY
	 * @see #FRIDAY
	 * @see #SATURDAY
	 * @see #SUNDAY
	 * @return A mask for the day of the week.
	 */
	public char GetDayOfWeekMask()
	{
		return dayOfWeekMask;
	}
	
	/**
	 * A mean of transportation.
	 * @author Piotrek
	 */
	private int meanOfTransportation;
	
	/**
	 * Gets the mean of transportation.
	 * @author Piotrek
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @return A mean of transportation.
	 */
	public int GetMeanOfTransportation()
	{
		return meanOfTransportation;
	}
	
	/**
	 * A list of stations in this connection.
	 * @author Piotrek
	 */
	private ArrayList stations = null;
	
	/**
	 * Gets a list of stations.
	 * @author Piotrek
	 * @return A list of stations.
	 */
	public ArrayList GetStations()
	{
		return stations;
	}
	
	/**
	 * A list of arrivals for all the stations in this connection.
	 * @author Piotrek
	 */
	private ArrayList arrivals = null;
	
	/**
	 * Gets a list of arrivals for all the stations.
	 * @author Piotrek
	 * @return A list of arrivals for all the stations.
	 */
	public ArrayList GetArrivals()
	{
		return arrivals;
	}
	
	/**
	 * A list of departures for all the stations in this connection.
	 * @author Piotrek
	 */
	private ArrayList departures = null;
	
	/**
	 * Gets a list of departures for all the stations.
	 * @author Piotrek
	 * @return A list of departures for all the stations.
	 */
	public ArrayList GetDepartures()
	{
		return departures;
	}
	
	/**
	 * Gets an index of the given station within the stations list of this connection.
	 * @author Piotrek
	 * @param station A station to search the index for.
	 * @return An index on the list of stations of the given station or -1 if the given station does not belong to this connection.
	 */
	public int GetStationIndex(Station station)
	{
		if (station == null)
			throw new NullPointerException("Station must not be null in Connection.GetStationIndex().");
		
		if (station.GetMeanOfTransportation() == meanOfTransportation)
			return stations.indexOf(station);
		
		throw new IllegalArgumentException("The given station does not support the mean of transportation this connection does.");
	}
	
	/**
	 * Gets an arrival time for the given station's index.
	 * @author Piotrek
	 * @param stationIndex A station's index.
	 * @return A time of arrival in minutes. It might be more then 24 hours (it's the next day of week in such a case).
	 */
	public int GetArrivalTime(int stationIndex)
	{
		if (stationIndex < 0 || stationIndex >= arrivals.size())
			throw new IndexOutOfBoundsException("Station's index out of range in Connection.GetArrivalTime().");
		
		Integer time = (Integer)arrivals.get(stationIndex);
		
		return time.intValue();
	}
	
	/**
	 * Gets a departure time for the given station's index.
	 * @author Piotrek
	 * @param stationIndex A station's index.
	 * @return A time of departure in minutes. It might be more then 24 hours (it's the next day of week in such a case).
	 */
	public int GetDepartureTime(int stationIndex)
	{
		if (stationIndex < 0 || stationIndex >= departures.size())
			throw new IndexOutOfBoundsException("Station's index out of range in Connection.GetDepartureTime().");
		
		Integer time = (Integer)departures.get(stationIndex);
		
		return time.intValue();
	}
	
	/**
	 * Converts the day of week constant from the Calendar class to the Connection class constant.
	 * @author Piotrek
	 * @param calendarDayOfWeek A constant from the Calendar class.
	 * @return A day of week Connection constant matching the given Calendar constant.
	 * @see Calendar#MONDAY
	 * @see Calendar#TUESDAY
	 * @see Calendar#WEDNESDAY
	 * @see Calendar#THURSDAY
	 * @see Calendar#FRIDAY
	 * @see Calendar#SATURDAY
	 * @see Calendar#SUNDAY  
	 */
	public static char CalendarToConnectionDayOfWeek(int calendarDayOfWeek)
	{
		switch (calendarDayOfWeek)
		{
			case Calendar.MONDAY:
				
				return MONDAY;
			
			case Calendar.TUESDAY:
				
				return TUESDAY;
			
			case Calendar.WEDNESDAY:
				
				return WEDNESDAY;
			
			case Calendar.THURSDAY:
				
				return THURSDAY;
			
			case Calendar.FRIDAY:
				
				return FRIDAY;
			
			case Calendar.SATURDAY:
				
				return SATURDAY;
			
			case Calendar.SUNDAY:
				
				return SUNDAY;
			
			default:
				
				throw new IllegalArgumentException("Unexpected day of week in Connection.CalendarToConnectionDayOfWeek().");
		}
	}
	
	/**
	 * Shifts the given day of week by the given number of days. If the shift is zero, the parameter dayOfWeek is returned as it is without checking if it is a valid day of week.
	 * @author Piotrek
	 * @param dayOfWeek A day of week to be shifted.
	 * @param shift A number of days to shift the given day of week.
	 * @return A shifted day of week.
	 * @see #MONDAY
	 * @see #TUESDAY
	 * @see #WEDNESDAY
	 * @see #THURSDAY
	 * @see #FRIDAY
	 * @see #SATURDAY
	 * @see #SUNDAY
	 */
	public static char ShiftDayOfWeek(char dayOfWeek, int shift)
	{
		shift %= 7;
		
		if (shift < 0)
			shift += 7;
		
		for (int i = 0; i < shift; ++i)
		{
			switch (dayOfWeek)
			{
				case MONDAY:
					
					dayOfWeek = TUESDAY;
					
					break;
					
				case TUESDAY:
					
					dayOfWeek = WEDNESDAY;
					
					break;
					
				case WEDNESDAY:
					
					dayOfWeek = THURSDAY;
					
					break;
					
				case THURSDAY:
					
					dayOfWeek = FRIDAY;
					
					break;
					
				case FRIDAY:
					
					dayOfWeek = SATURDAY;
					
					break;
					
				case SATURDAY:
					
					dayOfWeek = SUNDAY;
					
					break;
					
				case SUNDAY:
					
					dayOfWeek = MONDAY;
					
					break;
					
				default:
					
					throw new IllegalArgumentException("Unexpected day of week in Connection.ShiftDayOfWeek().");
			}
		}
		
		return dayOfWeek;
	}
	
	/**
	 * Gets the number of stations in this connection.
	 * @author Piotrek
	 * @return A number of stations in this connection.
	 */
	public int GetStationsCount()
	{
		return stations.size();
	}
	
	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 * @author Piotrek
	 */
	public boolean equals(Object other)
	{
		if (!Connection.class.isInstance(other))
			return false;
		
		Connection otherConnection = (Connection)other;
		
		// simple tests
		
		if (stations.size() != otherConnection.GetStationsCount())
			return false;
		
		if (meanOfTransportation != otherConnection.GetMeanOfTransportation())
			return false;
		
		if (dayOfWeekMask != otherConnection.GetDayOfWeekMask())
			return false;
		
		// checking times
		
		for (int i = 0; i < stations.size(); ++i)
		{
			if (GetArrivalTime((i)) != otherConnection.GetArrivalTime(i))
				return false;
			
			if (GetDepartureTime((i)) != otherConnection.GetDepartureTime(i))
				return false;
		}
		
		// comparaing stations
		
		Iterator myIterator = stations.iterator();
		Iterator otherIterator = otherConnection.GetStations().iterator();
		
		while (myIterator.hasNext()) // they have the same number of stations
		{
			Station myStation = (Station)myIterator.next();
			Station otherStation = (Station)otherIterator.next();
			
			if (!myStation.IsEqual(otherStation))
				return false;
		}
		
		return true;
	}
	
	/**
	 * Compares two connections.
	 * @author Piotrek
	 * @param other A connection this connection is to be compared against.
	 * @return A boolean value indicating whether both connections are equal.
	 */
	public boolean IsEqual(Connection other)
	{
		return equals(other);
	}
}
