/*
 * Created on 2005-03-29
 */
package world;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import javax.swing.JOptionPane;


/**
 * The main object - the world. It contains all other objects belonging to it.
 * @author Piotrek
 */
public final class World
{
	/**
	 * Constructor. Use @see #GetInstance() method to get and instance of this class.
	 * @author Piotrek
	 */
	private World()
	{
		// there is nothing to be done
	}
	
	/**
	 * The one and only instance of this class.
	 * @author Piotrek
	 */
	private static World world = null;
	
	/**
	 * A number assigned to the next air connection being added to the world.
	 * @author Piotrek
	 */
	private int nextAirConnectionNumber = 1;
		
	/**
	 * A number assigned to the next railroad connection being added to the world.
	 * @author Piotrek
	 */
	private int nextRailroadConnectionNumber = 1;
		
	/**
	 * A number assigned to the next bus connection being added to the world.
	 * @author Piotrek
	 */
	private int nextBusConnectionNumber = 1;
		
	/**
	 * A number assigned to the next sea connection being added to the world.
	 * @author Piotrek
	 */
	private int nextSeaConnectionNumber = 1;
		
	/**
	 * Gets the one and only instance of this class.
	 * @author Piotrek
	 * @return The only existing instance of this class.
	 */
	public static World GetInstance()
	{
		if (world == null)
			world = new World();
		
		return world;
	}
	
	/**
	 * Clears the world. It creates a new instance of the world, so the next call to the GetInstance() will return a clear world. 
	 * @author Piotrek
	 */
	public static void Clear()
	{
		world = new World();
	}
	
	/**
	 * A set of cities on this world.
	 * @author Piotrek
	 */
	private ArrayList cities = new ArrayList(50);
	
	/**
	 * Adds a single city to the list of cities on this world.
	 * @author Piotrek
	 * @param city A city to be added.
	 */
	public void AddCity(City city)
	{
		if (city == null)
			throw new NullPointerException("City must not be null in World.AddCity().");
		
		cities.add(city);
	}
	
	/**
	 * Removes a given city from this world.
	 * @author Piotrek
	 * @param city A city to be removed.
	 */
	public void RemoveCity(City city)
	{
		if (city == null)
			throw new NullPointerException("City must not be null in World.RemoveCity().");
		
		cities.remove(city);
	}
	
	/**
	 * Gets an iterator to the cities on this world.
	 * @author Piotrek
	 * @return The iterator to the cities on this world.
	 */
	public Iterator GetCities()
	{
		return cities.iterator();
	}
	
	/**
	 * A set of airports on this world.
	 * @author Piotrek
	 */
	private ArrayList airports = new ArrayList(20);
	
	/**
	 * Adds a single airport to the list of airports on this world.
	 * @author Piotrek
	 * @param airport An airport to be added.
	 */
	public void AddAirport(Airport airport)
	{
		if (airport == null)
			throw new NullPointerException("Airport must not be null in World.AddAirport().");
		
		airports.add(airport);
	}
	
	/**
	 * Removes a given airport from this world.
	 * @author Piotrek
	 * @param airport An airport to be removed.
	 */
	public void RemoveAirport(Airport airport)
	{
		if (airport == null)
			throw new NullPointerException("Airport must not be null in World.RemoveAirport().");
		
		airports.remove(airport);
		
		RemoveConnectionsWithStation(airport);
	}
	
	/**
	 * Gets an iterator to the airports on this world.
	 * @author Piotrek
	 * @return The iterator to the airports on this world.
	 */
	public Iterator GetAirports()
	{
		return airports.iterator();
	}
	
	/**
	 * A set of train stations on this world.
	 * @author Piotrek
	 */
	private ArrayList trainStations = new ArrayList(50);
	
	/**
	 * Adds a single train station to the list of train stations on this world.
	 * @author Piotrek
	 * @param trainStation A train station to be added.
	 */
	public void AddTrainStation(TrainStation trainStation)
	{
		if (trainStation == null)
			throw new NullPointerException("Train station must not be null in World.AddTrainStation().");
		
		trainStations.add(trainStation);
	}
	
	/**
	 * Removes a given train station from this world.
	 * @author Piotrek
	 * @param trainStation A train station to be removed.
	 */
	public void RemoveTrainStation(TrainStation trainStation)
	{
		if (trainStation == null)
			throw new NullPointerException("Train station must not be null in World.RemoveTrainStation().");
		
		trainStations.remove(trainStation);
		
		RemoveConnectionsWithStation(trainStation);
	}
	
	/**
	 * Gets an iterator to the train stations on this world.
	 * @author Piotrek
	 * @return The iterator to the train stations on this world.
	 */
	public Iterator GetTrainStations()
	{
		return trainStations.iterator();
	}
	
	/**
	 * A set of bus stops on this world.
	 * @author Piotrek
	 */
	private ArrayList busStops = new ArrayList(50);
	
	/**
	 * Adds a single bus stop to the list of bus stops on this world.
	 * @author Piotrek
	 * @param busStop A bus stop to be added.
	 */
	public void AddBusStop(BusStop busStop)
	{
		if (busStop == null)
			throw new NullPointerException("Bus stop must not be null in World.AddBusStation().");
		
		busStops.add(busStop);
	}
	
	/**
	 * Removes a given bus stop from this world.
	 * @author Piotrek
	 * @param busStop A bus stop to be removed.
	 */
	public void RemoveBusStop(BusStop busStop)
	{
		if (busStop == null)
			throw new NullPointerException("Bus stop must not be null in World.RemoveBusStation().");
		
		busStops.remove(busStop);
		
		RemoveConnectionsWithStation(busStop);
	}
	
	/**
	 * Gets an iterator to the bus stops on this world.
	 * @author Piotrek
	 * @return The iterator to the bus stops on this world.
	 */
	public Iterator GetBusStops()
	{
		return busStops.iterator();
	}
	
	/**
	 * A set of harbours on this world.
	 * @author Piotrek
	 */
	private ArrayList harbours = new ArrayList(20);
	
	/**
	 * Adds a single harbour to the list of harbours on this world.
	 * @author Piotrek
	 * @param harbour A harbour to be added.
	 */
	public void AddHarbour(Harbour harbour)
	{
		if (harbour == null)
			throw new NullPointerException("Harbour must not be null in World.AddHarbour().");
		
		harbours.add(harbour);
	}
	
	/**
	 * Removes a given harbour from this world.
	 * @author Piotrek
	 * @param harbour A harbour to be removed.
	 */
	public void RemoveHarbour(Harbour harbour)
	{
		if (harbour == null)
			throw new NullPointerException("Harbour must not be null in World.RemoveHarbour().");
		
		harbours.remove(harbour);
		
		RemoveConnectionsWithStation(harbour);
	}
	
	/**
	 * Gets an iterator to the harbours on this world.
	 * @author Piotrek
	 * @return The iterator to the harbours on this world.
	 */
	public Iterator GetHarbours()
	{
		return harbours.iterator();
	}
	
	/**
	 * Removes all the connections that contain the given station.
	 * @author Piotrek
	 * @param station All the connections containing this station will be removed.
	 */
	private void RemoveConnectionsWithStation(Station station)
	{
		if (station == null)
			throw new NullPointerException("Station must not be null in World.RemoveConnectionsWithStation().");
		
		// building a list of connections to be removed
		
		ArrayList connectionsToBeRemoved = new ArrayList();
		
		Iterator iterator = connections.iterator();
		
		while (iterator.hasNext())
		{
			Connection connection = (Connection)iterator.next();
			
			if (connection.GetStations().contains(station))
				connectionsToBeRemoved.add(connection);
		}
		
		// removing them
		
		iterator = connectionsToBeRemoved.iterator();
		
		while (iterator.hasNext())
			connections.remove(iterator.next());
	}
	
	/**
	 * A list of connections.
	 * @author Piotrek
	 */
	private ArrayList connections = new ArrayList(50);
	
	/**
	 * Add a new connection to this world.
	 * @author Piotrek
	 * @param connection
	 */
	public void AddConnection(Connection connection)
	{
		if (connection == null)
			throw new NullPointerException("A connection in World.AddConnection() is null.");
		
		// checking name and updating a proper connection number
		
		switch (connection.GetMeanOfTransportation())
		{
			// flight
			case MeanOfTransportation.AIRPLANE:	
			{
				String flightName = connection.GetName();
			
				if (!flightName.startsWith("Flight #"))
					throw new IllegalArgumentException("A flight connection's name does not start with `Flight #' in World.AddConnection().");
				
				flightName = flightName.substring(8);
				
				int flightNumber = 0;
				
				try
				{
					flightNumber = Integer.parseInt(flightName);
				}
				
				catch (NumberFormatException xcp)
				{
					throw new IllegalArgumentException("Error while parsing the flight number in World.AddConnection().");
				}
				
				if (nextAirConnectionNumber - 1 > flightNumber)
					throw new IllegalArgumentException("The flight number is not a valid one in World.AddConnection(). Use World.GetNextConnectionName() to get a valid name.");
				
				nextAirConnectionNumber = flightNumber + 1;
			}	
			
			break;
				
			// bus
			case MeanOfTransportation.BUS:
			{
				String busName = connection.GetName();
			
				if (!busName.startsWith("Bus #"))
					throw new IllegalArgumentException("A bus connection's name does not start with `Bus #' in World.AddConnection().");
				
				busName = busName.substring(5);
				
				int busNumber = 0;
				
				try
				{
					busNumber = Integer.parseInt(busName);
				}
				
				catch (NumberFormatException xcp)
				{
					throw new IllegalArgumentException("Error while parsing the bus number in World.AddConnection().");
				}
				
				if (nextBusConnectionNumber - 1 > busNumber)
					throw new IllegalArgumentException("The bus number is not a valid one in World.AddConnection(). Use World.GetNextConnectionName() to get a valid name.");
					
				nextBusConnectionNumber = busNumber + 1;
			}
			
			break;
			
			// train
			case MeanOfTransportation.TRAIN:
			{	
				String trainName = connection.GetName();
			
				if (!trainName.startsWith("Train #"))
					throw new IllegalArgumentException("A train connection's name does not start with `Train #' in World.AddConnection().");
				
				trainName = trainName.substring(7);
				
				int trainNumber = 0;
				
				try
				{
					trainNumber = Integer.parseInt(trainName);
				}
				
				catch (NumberFormatException xcp)
				{
					throw new IllegalArgumentException("Error while parsing the train number in World.AddConnection().");
				}
				
				if (nextRailroadConnectionNumber - 1 > trainNumber)
					throw new IllegalArgumentException("The train number is not a valid one in World.AddConnection(). Use World.GetNextConnectionName() to get a valid name.");
					
				nextRailroadConnectionNumber = trainNumber + 1;
			}
			
			break;
			
			// harbour
			case MeanOfTransportation.SHIP:
			{	
				String cruiseName = connection.GetName();
			
				if (!cruiseName.startsWith("Cruise #"))
					throw new IllegalArgumentException("A sea connection's name does not start with `Cruise #' in World.AddConnection().");
				
				cruiseName = cruiseName.substring(8);
				
				int cruiseNumber = 0;
				
				try
				{
					cruiseNumber = Integer.parseInt(cruiseName);
				}
				
				catch (NumberFormatException xcp)
				{
					throw new IllegalArgumentException("Error while parsing the cruise number in World.AddConnection().");
				}
				
				if (nextSeaConnectionNumber - 1 > cruiseNumber)
					throw new IllegalArgumentException("The cruise number is not a valid one in World.AddConnection(). Use World.GetNextConnectionName() to get a valid name.");
					
				nextSeaConnectionNumber = cruiseNumber + 1;
			}
			
			break;
		}
		
		// checking the stations
		
		Iterator iterator = connection.GetStations().iterator();
		
		while (iterator.hasNext())
		{
			WorldObject station = (WorldObject)iterator.next();
			
			switch (connection.GetMeanOfTransportation())
			{
				case MeanOfTransportation.AIRPLANE:
					
					if (!airports.contains(station))
						throw new IllegalArgumentException("A connection in World.AddConnetion() contains a station out of this world.");
					
					break;
					
				case MeanOfTransportation.BUS:
					
					if (!busStops.contains(station))
						throw new IllegalArgumentException("A connection in World.AddConnetion() contains a station out of this world.");
					
					break;
					
				case MeanOfTransportation.SHIP:
					
					if (!harbours.contains(station))
						throw new IllegalArgumentException("A connection in World.AddConnetion() contains a station out of this world.");
					
					break;
					
				case MeanOfTransportation.TRAIN:
					
					if (!trainStations.contains(station))
						throw new IllegalArgumentException("A connection in World.AddConnetion() contains a station out of this world.");
										
					break;
					
				default:
					
					throw new IllegalArgumentException("A station on the connection's list is of the unexpected type in World.AddConnection().");
			}
		}
				
		// it's all OK, so we can add it
		
		connections.add(connection);
	}
	
	/**
	 * Removes a given connection from this world.
	 * @author Piotrek
	 * @param connection A connection to be removed.
	 */
	public void RemoveConnection(Connection connection)
	{
		if (connection == null)
			throw new NullPointerException("Connection must not be null in World.RemoveConnection().");
		
		connections.remove(connection);
	}
	
	/**
	 * Gets an iterator for the connections.
	 * @author Piotrek
	 * @return An iterator for the connections.
	 */
	public Iterator GetConnections()
	{
		return connections.iterator();
	}
	
	/**
	 * Saves this world to the given file.
	 * @author Piotrek
	 * @param fileName A name of the file to save the world to.
	 */
	public void SaveAs(String fileName)
	{
		if (fileName == null)
			throw new NullPointerException("File name must not be null in World.SaveAs().");
		
		try
		{
			int stationID = 0;
			
			HashMap stationToIDMap = new HashMap();
			
			FileWriter fileWriter = new FileWriter(fileName);
			
			fileWriter.write("<?xml version=\"1.0\"?>\r\n");
			
			fileWriter.write("<World>\r\n");
			
			// cities
			fileWriter.write("   <Cities>\r\n");
			
			Iterator iterator = cities.iterator();
			
			while (iterator.hasNext())
			{
				WorldObject city = (WorldObject)iterator.next();
				
				fileWriter.write("      <City name=\"" + city.name + "\">\r\n");
				
				SaveLocationTag(fileWriter, city.GetLocation(), "         ");
				
				fileWriter.write("      </City>\r\n");
			}
			
			fileWriter.write("   </Cities>\r\n");
			
			// airports
			fileWriter.write("   <Airports>\r\n");
			
			iterator = airports.iterator();
			
			while (iterator.hasNext())
			{
				WorldObject airport = (WorldObject)iterator.next();
				
				fileWriter.write("      <Airport name=\"" + airport.name + "\" id=\"" + stationID + "\">\r\n");
				
				SaveLocationTag(fileWriter, airport.GetLocation(), "         ");
				
				fileWriter.write("      </Airport>\r\n");
				
				stationToIDMap.put(airport, new Integer(stationID++));
			}
			
			fileWriter.write("   </Airports>\r\n");
			
			// bus stops
			fileWriter.write("   <BusStops>\r\n");
			
			iterator = busStops.iterator();
			
			while (iterator.hasNext())
			{
				WorldObject busStop = (WorldObject)iterator.next();
				
				fileWriter.write("      <BusStop name=\"" + busStop.name + "\" id=\"" + stationID + "\">\r\n");
				
				SaveLocationTag(fileWriter, busStop.GetLocation(), "         ");
				
				fileWriter.write("      </BusStop>\r\n");
				
				stationToIDMap.put(busStop, new Integer(stationID++));
			}
			
			fileWriter.write("   </BusStops>\r\n");
			
			// train stations
			fileWriter.write("   <TrainStations>\r\n");
			
			iterator = trainStations.iterator();
			
			while (iterator.hasNext())
			{
				WorldObject trainStation = (WorldObject)iterator.next();
				
				fileWriter.write("      <TrainStation name=\"" + trainStation.name + "\" id=\"" + stationID + "\">\r\n");
				
				SaveLocationTag(fileWriter, trainStation.GetLocation(), "         ");
				
				fileWriter.write("      </TrainStation>\r\n");
				
				stationToIDMap.put(trainStation, new Integer(stationID++));
			}
			
			fileWriter.write("   </TrainStations>\r\n");
			
			// harbours
			
			fileWriter.write("   <Harbours>\r\n");
			
			iterator = harbours.iterator();
			
			while (iterator.hasNext())
			{
				WorldObject harbour = (WorldObject)iterator.next();
				
				fileWriter.write("      <Harbour name=\"" + harbour.name + "\" id=\"" + stationID + "\">\r\n");
				
				SaveLocationTag(fileWriter, harbour.GetLocation(), "         ");
				
				fileWriter.write("      </Harbour>\r\n");
				
				stationToIDMap.put(harbour, new Integer(stationID++));
			}
			
			fileWriter.write("   </Harbours>\r\n");
			
			// connections
			fileWriter.write("   <Connections>\r\n");
			
			iterator = connections.iterator();
			
			while (iterator.hasNext())
			{
				Connection connection = (Connection)iterator.next();
				
				fileWriter.write("      <Connection name=\"" + connection.GetName() + "\" meanOfTransportation=\"" + MeanOfTransportation.GetMeanOfTransportationString(connection.GetMeanOfTransportation()) + "\">\r\n");
								
				// stations
				
				fileWriter.write("         <StationIDs>\r\n");
				
				Iterator stations = connection.GetStations().iterator();
				
				while (stations.hasNext())
				{
					WorldObject station = (WorldObject)stations.next();
					
					int id = ((Integer)stationToIDMap.get(station)).intValue();
					
					fileWriter.write("            <StationID id=\"" + id + "\"/>\r\n");
					
				}
				
				fileWriter.write("         </StationIDs>\r\n");
				
				// arrival times
				
				fileWriter.write("         <ArrivalTimes>\r\n");
				
				Iterator arrivals = connection.GetArrivals().iterator();
				
				while (arrivals.hasNext())
				{
					int arrivalTime = ((Integer)arrivals.next()).intValue();
										
					fileWriter.write("            <ArrivalTime time=\"" + TravelTime.GetTime(arrivalTime) + "\"/>\r\n");
				}
				
				fileWriter.write("         </ArrivalTimes>\r\n");
				
				// departures
				
				fileWriter.write("         <DepartureTimes>\r\n");
				
				Iterator departures = connection.GetDepartures().iterator();
				
				while (departures.hasNext())
				{
					int departureTime = ((Integer)departures.next()).intValue();
										
					fileWriter.write("            <DepartureTime time=\"" + TravelTime.GetTime(departureTime) + "\"/>\r\n");
				}
				
				fileWriter.write("         </DepartureTimes>\r\n");
				
				// days of week
				
				fileWriter.write("         <DaysOfWeek>\r\n");
				
				int dayOfWeekMask = connection.GetDayOfWeekMask();
				
				if ((dayOfWeekMask & Connection.MONDAY) > 0)
					fileWriter.write("            <Monday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Monday exists=\"false\"/>\r\n");
				
				if ((dayOfWeekMask & Connection.TUESDAY) > 0)
					fileWriter.write("            <Tuesday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Tuesday exists=\"false\"/>\r\n");
				
				if ((dayOfWeekMask & Connection.WEDNESDAY) > 0)
					fileWriter.write("            <Wednesday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Wednesday exists=\"false\"/>\r\n");
				
				if ((dayOfWeekMask & Connection.THURSDAY) > 0)
					fileWriter.write("            <Thursday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Thursday exists=\"false\"/>\r\n");
				
				if ((dayOfWeekMask & Connection.FRIDAY) > 0)
					fileWriter.write("            <Friday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Friday exists=\"false\"/>\r\n");
				
				if ((dayOfWeekMask & Connection.SATURDAY) > 0)
					fileWriter.write("            <Saturday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Saturday exists=\"false\"/>\r\n");
				
				if ((dayOfWeekMask & Connection.SUNDAY) > 0)
					fileWriter.write("            <Sunday exists=\"true\"/>\r\n");
				else
					fileWriter.write("            <Sunday exists=\"false\"/>\r\n");
				
				fileWriter.write("         </DaysOfWeek>\r\n");
				
				fileWriter.write("      </Connection>\r\n");
			}
			
			fileWriter.write("   </Connections>\r\n");

			// closing the world and a whole file
			fileWriter.write("</World>\r\n");
			
			fileWriter.flush();
			
			fileWriter.close();
		}
		
		catch (IOException xcp)
		{
			JOptionPane.showMessageDialog(null, "There was an exception while saving to the file of  `" + fileName + "': " + xcp.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		}
	}
	
	/**
	 * Stores a location tag for the given location object to the given file writer. 
	 * @author Piotrek
	 * @param fileWriter A file writer to write to.
	 * @param location A location to be serialised.
	 * @param linePrefix A string added to each line (Spaces).
	 * @throws IOException Thrown in case of error.
	 */
	private static void SaveLocationTag(FileWriter fileWriter, Location location, String linePrefix) throws IOException
	{
		if (fileWriter == null)
			throw new NullPointerException("File writer must be not null in World.SaveLocationTag().");
		
		if (location == null)
			throw new NullPointerException("Location must be not null in World.SaveLocationTag().");
		
		if (linePrefix == null)
			throw new NullPointerException("Line prefix must be not null in World.SaveLocationTag().");
		
		fileWriter.write(linePrefix + "<Location>\r\n");
		
		fileWriter.write(linePrefix + "   <X x=\"" + location.x + "\"/>\r\n");
		fileWriter.write(linePrefix + "   <Y y=\"" + location.y + "\"/>\r\n");
		fileWriter.write(linePrefix + "   <ContinentName name=\"" + location.continentName + "\"/>\r\n");
		fileWriter.write(linePrefix + "   <IslandName name=\"" + location.islandName + "\"/>\r\n");
		
		fileWriter.write(linePrefix + "</Location>\r\n");
	}
	
	/**
	 * Gets a unique name for the connection for the given mean of transportation.
	 * @author Piotrek
	 * @param meanOfTransportation
	 * @return A unique name for the connection for the given mean of transportation.
	 */
	public String GetNextConnectionName(int meanOfTransportation)
	{
		switch (meanOfTransportation)
		{
			case MeanOfTransportation.AIRPLANE:
				
				return "Flight #" + nextAirConnectionNumber++;
							
			case MeanOfTransportation.BUS:
				
				return "Bus #" + nextBusConnectionNumber++;
				
			case MeanOfTransportation.SHIP:
				
				return "Cruise #" + nextSeaConnectionNumber++;
				
			case MeanOfTransportation.TRAIN:
				
				return "Train #" + nextRailroadConnectionNumber++;
			
			default:
				
				throw new IllegalArgumentException("Unexpected mean of transportation in WorldGetNextConnectionName().");
		}
	}
}
