/*
 * Created on 2005-04-15
 */
package world;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A parser for world files.
 * @author Piotrek
 */
public final class WorldFileParser extends DefaultHandler
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public WorldFileParser()
	{
		super();
	}
	
	/**
	 * Parses the given file. After a call to this method a world singleton contains 
	 * a world stored in the given file. The world is changed only if the file has been
	 * parsed without any errors.
	 * @author Piotrek
	 * @param fileName The name of the file to be parsed.
	 * @throws Exception Thrown in case of error.
	 * @throws IOException In case of any IO error.
	 * @throws SAXException In case of some parsing error.
	 */
	public void Parse(String fileName) throws Exception, IOException, SAXException
	{
		if (fileName == null)
			throw new NullPointerException("File name must not be null in WorldFileParser.Parse().");
			
		ClearData();
				
		SAXParser saxParser;
		
		try
		{
			saxParser = CreateSAXParser();
		}
		
		catch (Throwable xcp)
		{
			throw new Exception("Failed to create a parser.");
		}
		
		saxParser.parse(new File(fileName), this);
		
		ChangeWorld();
	}
	
	/**
	 * Clears all the internal data, so the parser is ready to parse a new file.
	 * @author Piotrek
	 */
	private void ClearData()
	{
		cities.clear();
		connections.clear();
		stations.clear();
		usedConnectionNames.clear();
		usedIDs.clear();
		idToStationMap.clear();
		
		fInAirport = false;
		fInAirports = false;
		fInArrivalTimes = false;
		fInBusStop = false;
		fInBusStops = false;
		fInCities = false;
		fInCity = false;
		fInConnection = false;
		fInConnections = false;
		fInDaysOfWeek = false;
		fInDepartureTimes = false;
		fInHarbour = false;
		fInHarbours = false;
		fInLocation = false;
		fInStationIDs = false;
		fInTrainStation = false;
		fInTrainStations = false;
	}
		
	/**
	 * Creates an instance of the SAXParser.
	 * @author Piotrek
	 * @return The requested parser.
	 * @throws ParserConfigurationException If it was imposible to create a parser satisfing the current configuration.
	 * @throws SAXException
	 */
	private static SAXParser CreateSAXParser() throws ParserConfigurationException, SAXException
	{
	 	SAXParserFactory factory = SAXParserFactory.newInstance();
		
	 	SAXParser saxParser = factory.newSAXParser();
	 	
	 	return saxParser;
	}
	
	/**
	 * Indicates whether the parser is inside a `world' tag.
	 * @author Piotrek
	 */
	private boolean fInWorld = false;
	
	/**
	 * Indicates whether the parser is inside a `cities' tag.
	 * @author Piotrek
	 */
	private boolean fInCities = false;
	
	/**
	 * Indicates whether the parser is inside a `city' tag.
	 * @author Piotrek
	 */
	private boolean fInCity = false;
	
	/**
	 * Indicates whether the parser is inside a `location' tag.
	 * @author Piotrek
	 */
	private boolean fInLocation = false;
	
	/**
	 * Indicates whether the parser is inside an `airports' tag.
	 * @author Piotrek
	 */
	private boolean fInAirports = false;
	
	/**
	 * Indicates whether the parser is inside an `airport' tag.
	 * @author Piotrek
	 */
	private boolean fInAirport = false;
	
	/**
	 * Indicates whether the parser is inside a `bus stops' tag.
	 * @author Piotrek
	 */
	private boolean fInBusStops = false;
	
	/**
	 * Indicates whether the parser is inside a `bus stop' tag.
	 * @author Piotrek
	 */
	private boolean fInBusStop = false;
	
	/**
	 * Indicates whether the parser is inside a `train stations' tag.
	 * @author Piotrek
	 */
	private boolean fInTrainStations = false;
	
	/**
	 * Indicates whether the parser is inside a `train station' tag.
	 * @author Piotrek
	 */
	private boolean fInTrainStation = false;
	
	/**
	 * Indicates whether the parser is inside a `harbours' tag.
	 * @author Piotrek
	 */
	private boolean fInHarbours = false;
	
	/**
	 * Indicates whether the parser is inside a `harbour' tag.
	 * @author Piotrek
	 */
	private boolean fInHarbour = false;
	
	/**
	 * Indicates whether the parser is inside a `connections' tag.
	 * @author Piotrek
	 */
	private boolean fInConnections = false;
	
	/**
	 * Indicates whether the parser is inside a `connection' tag.
	 * @author Piotrek
	 */
	private boolean fInConnection = false;
	
	/**
	 * Indicates whether the parser is inside a `stations ids' tag.
	 * @author Piotrek
	 */
	private boolean fInStationIDs = false;
	
	/**
	 * Indicates whether the parser is inside a `arrival times' tag.
	 * @author Piotrek
	 */
	private boolean fInArrivalTimes = false;
	
	/**
	 * Indicates whether the parser is inside a `departure times' tag.
	 * @author Piotrek
	 */
	private boolean fInDepartureTimes = false;
	
	/**
	 * Indicates whether the parser is inside a `days of week' tag.
	 * @author Piotrek
	 */
	private boolean fInDaysOfWeek = false;
	
	/**
	 * A name of a city. Valid only if in a city tag.
	 * @author Piotrek
	 */
	private String cityName = null;
	
	/**
	 * A name of the airport. Valid only if in an airport tag.
	 * @author Piotrek
	 */
	private String airportName = null;
	
	/**
	 * An ID of the current airport. Valid only if in an airport tag.
	 * @author Piotrek
	 */
	private int airportID = 0;
	
	/**
	 * A name of the bus stop. Valid only if in a bus stop tag.
	 * @author Piotrek
	 */
	private String busStopName = null;
	
	/**
	 * An ID of the current bus stop. Valid only if in a bus stop tag.
	 * @author Piotrek
	 */
	private int busStopID = 0;
	
	/**
	 * A name of the train station. Valid only if in a train station tag.
	 * @author Piotrek
	 */
	private String trainStationName = null;
	
	/**
	 * An ID of the current train station. Valid only if in a train station tag.
	 * @author Piotrek
	 */
	private int trainStationID = 0;
	
	/**
	 * A name of the harbour. Valid only if in a harbour tag.
	 * @author Piotrek
	 */
	private String harbourName = null;
	
	/**
	 * An ID of the current harbour. Valid only if in a harbour tag.
	 * @author Piotrek
	 */
	private int harbourID = 0;
	
	/**
	 * A name of the connection.
	 * @author Piotrek
	 */
	private String connectionName = null;
	
	/**
	 * A mean of transportation for the connection.
	 * @author Piotrek 
	 */
	private int connectionMeanOfTransportation;
	
	/**
	 * A list of used IDs for the stations. Used to verify if the new ID is unique.
	 * @author Piotrek
	 */
	private ArrayList usedIDs = new ArrayList();
	
	/**
	 * A list of used connection names. Used to verify if the new connection name is unique.
	 * @author Piotrek
	 */
	private ArrayList usedConnectionNames = new ArrayList();
	
	/**
	 * An ID of the station assigned to the connection.
	 * @author Piotrek
	 */
	private int stationID = -1;
	
	/**
	 * A maping from tha station ID to the station.
	 * @author Piotrek
	 */
	private HashMap idToStationMap = new HashMap();
	
	/**
	 * A list of all parsed stations.
	 * @author Piotrek
	 */
	private ArrayList stations = new ArrayList();
	
	/**
	 * An arrival time for the station assigned to the connection.
	 * @author Piotrek
	 */
	private int arrivalTime = -1;
	
	/**
	 * A departure time for the station assigned to the connection.
	 * @author Piotrek
	 */
	private int departureTime = -1;
	
	/**
	 * A day of the week flag for the connection.
	 * @author Piotrek
	 */
	private char dayOfWeekFlag = 0;
	
	/**
	 * The x coordinate of the location.
	 * @author Piotrek
	 */
	private int x = -1;
	
	/**
	 * The y coordinate of the location.
	 * @author Piotrek
	 */
	private int y = -1;
	
	/**
	 * The continentName name in the location.
	 * @author Piotrek
	 */
	private String continentName = null;
	
	/**
	 * The the island name in the location.
	 * @author Piotrek
	 */
	private String islandName = null;
	
	/**
	 * A list of all parsed cities.
	 * @author Piotrek
	 */
	private ArrayList cities = new ArrayList();
	
	/**
	 * A list of stations assigned to the current connection.
	 * @author Piotrek
	 */
	private ArrayList connectionStations = new ArrayList();
	
	/**
	 * A list of arrival times for the current connection.
	 * @author Piotrek
	 */
	private ArrayList connectionArrivalTimes = new ArrayList();
	
	/**
	 * A list of departure times for the current connection.
	 * @author Piotrek
	 */
	private ArrayList connectionDepartureTimes = new ArrayList();
	
	/**
	 * A list of all parsed connections.
	 * @author Piotrek
	 */
	private ArrayList connections = new ArrayList();
		
	/**
	 * @author Piotrek
	 * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uriNamespace, String simpleName, String qualifiedName, Attributes attributes) throws SAXException
	{
		String name = simpleName;
		
		if (name.compareTo("") == 0) 
			name = qualifiedName; // not namespace aware
		
		if (name.compareTo("World") == 0)
			fInWorld = true;
		
		// cities
		if (name.compareTo("Cities") == 0)
		{
			if (!fInWorld)
				throw new SAXParseException("Cities tag outside the World tag.", locator);
			
			fInCities = true;
		}
		
		if (name.compareTo("City") == 0)
		{			
			if (!fInCities)
				throw new SAXParseException("City tag outside the Cities tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("City tag has an invalid number of attributes.", locator);
			
			String cityName = attributes.getValue("name");
			
			if (cityName == null)
				throw new SAXParseException("City tag has no `name' attribute.", locator);
			
			this.cityName = cityName;
			
			fInCity = true;
			
			ClearLocation();
		}
		
		// airports
		if (name.compareTo("Airports") == 0)
		{
			if (!fInWorld)
				throw new SAXParseException("Airports tag outside the World tag.", locator);
			
			fInAirports = true;
		}
		
		if (name.compareTo("Airport") == 0)
		{
			if (!fInAirports)
				throw new SAXParseException("Airport tag outside the Airports tag.", locator);
			
			fInAirport = true;
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("Airport tag has an invalid number of attributes.", locator);
			
			String airportName = attributes.getValue("name");		
			
			if (airportName == null)
				throw new SAXParseException("Airport tag has no `name' attribute.", locator);
						
			String airportIDString = attributes.getValue("id");
			
			if (airportIDString == null)
				throw new SAXParseException("Airport tag has no `id' attribute.", locator);
			
			this.airportName = airportName;
			
			try
			{
				airportID = Integer.parseInt(airportIDString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException(xcp.getMessage(), locator);
			}
			
			if (usedIDs.contains(new Integer(airportID)))
				throw new SAXParseException("The airport has not got a unique ID.", locator);
			
			usedIDs.add(new Integer(airportID));
			
			ClearLocation();
		}
		
		// bus stops
		if (name.compareTo("BusStops") == 0)
		{
			if (!fInWorld)
				throw new SAXParseException("BusStops tag outside the World tag.", locator);
			
			fInBusStops = true;
		}
		
		if (name.compareTo("BusStop") == 0)
		{
			if (!fInBusStops)
				throw new SAXParseException("BusStop tag outside the BusStops tag.", locator);
			
			fInBusStop = true;
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("BusStop tag has an invalid number of attributes.", locator);
			
			String busStopName = attributes.getValue("name");		
			
			if (busStopName == null)
				throw new SAXParseException("BusStop tag has no `name' attribute.", locator);
						
			String busStopIDString = attributes.getValue("id");
			
			if (busStopIDString == null)
				throw new SAXParseException("BusStop tag has no `id' attribute.", locator);
			
			this.busStopName = busStopName;
			
			try
			{
				busStopID = Integer.parseInt(busStopIDString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException(xcp.getMessage(), locator);
			}
			
			if (usedIDs.contains(new Integer(busStopID)))
				throw new SAXParseException("The bus stop has not got a unique ID.", locator);
			
			usedIDs.add(new Integer(busStopID));
			
			ClearLocation();
		}
		
		// train station
		if (name.compareTo("TrainStations") == 0)
		{
			if (!fInWorld)
				throw new SAXParseException("TrainStations tag outside the World tag.", locator);
			
			fInTrainStations = true;
		}
		
		if (name.compareTo("TrainStation") == 0)
		{
			if (!fInTrainStations)
				throw new SAXParseException("TrainStation tag outside the TrainStations tag.", locator);
			
			fInTrainStation = true;
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("TrainStation tag has an invalid number of attributes.", locator);
			
			String trainStationName = attributes.getValue("name");		
			
			if (trainStationName == null)
				throw new SAXParseException("TrainStation tag has no `name' attribute.", locator);
						
			String trainStationIDString = attributes.getValue("id");
			
			if (trainStationIDString == null)
				throw new SAXParseException("TrainStation tag has no `id' attribute.", locator);
			
			this.trainStationName = trainStationName;
			
			try
			{
				trainStationID = Integer.parseInt(trainStationIDString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException(xcp.getMessage(), locator);
			}
			
			if (usedIDs.contains(new Integer(trainStationID)))
				throw new SAXParseException("The train station has not got a unique ID.", locator);
			
			usedIDs.add(new Integer(trainStationID));
			
			ClearLocation();
		}
		
		// harbours
		if (name.compareTo("Harbours") == 0)
		{
			if (!fInWorld)
				throw new SAXParseException("Harbours tag outside the World tag.", locator);
			
			fInHarbours = true;
		}
		
		if (name.compareTo("Harbour") == 0)
		{
			if (!fInHarbours)
				throw new SAXParseException("Harbour tag outside the Harbours tag.", locator);
			
			fInHarbour = true;
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("Harbour tag has an invalid number of attributes.", locator);
			
			String harbourName = attributes.getValue("name");		
			
			if (harbourName == null)
				throw new SAXParseException("Harbour tag has no `name' attribute.", locator);
						
			String harbourIDString = attributes.getValue("id");
			
			if (harbourIDString == null)
				throw new SAXParseException("Harbour tag has no `id' attribute.", locator);
			
			this.harbourName = harbourName;
			
			try
			{
				harbourID = Integer.parseInt(harbourIDString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException(xcp.getMessage(), locator);
			}
			
			if (usedIDs.contains(new Integer(harbourID)))
				throw new SAXParseException("The harbour has not got a unique ID.", locator);
			
			usedIDs.add(new Integer(harbourID));
			
			ClearLocation();
		}
		
		// connections
		if (name.compareTo("Connections") == 0)
		{
			if (!fInWorld)
				throw new SAXParseException("Connections tag outside the World tag.", locator);
			
			fInConnections = true;
		}
		
		if (name.compareTo("Connection") == 0)
		{
			if (!fInConnections)
				throw new SAXParseException("Connection tag outside the Connections tag.", locator);
			
			fInConnection = true;
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("Connection tag has an invalid number of attributes.", locator);
			
			String connectionName = attributes.getValue("name");		
			
			if (connectionName == null)
				throw new SAXParseException("Connection tag has no `name' attribute.", locator);
			
			if (usedConnectionNames.contains(connectionName))
				throw new SAXParseException("A connection has not a unique name assigned.", locator);
			
			usedConnectionNames.add(connectionName);
						
			String connectionMeanOfTransportationString = attributes.getValue("meanOfTransportation");
						
			if (connectionMeanOfTransportationString == null)
				throw new SAXParseException("Connection tag has no `meanOfTransportation' attribute.", locator);
			
			try
			{
				this.connectionMeanOfTransportation = MeanOfTransportation.GetMeanOfTransportationCode(connectionMeanOfTransportationString);
			}
			
			catch (IllegalArgumentException xcp)
			{
				throw new SAXParseException("The connection has an invalid mean of transportation assigned.", locator);
			}
			
			this.connectionName = connectionName;
			
			connectionStations = new ArrayList();
			connectionArrivalTimes = new ArrayList();
			connectionDepartureTimes = new ArrayList();
		}
		
		// station IDs
		if (name.compareTo("StationIDs") == 0)
		{
			if (!fInConnection)
				throw new SAXParseException("StationIDs tag outside the Connection tag.", locator);
			
			fInStationIDs = true;
		}
		
		if (name.compareTo("StationID") == 0)
		{
			if (!fInStationIDs)
				throw new SAXParseException("StationID tag outside the StationIDs tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("StationID tag has an invalid number of attributes.", locator);
						
			String stationIDString = attributes.getValue("id");
			
			if (stationIDString == null)
				throw new SAXParseException("StationID tag has no `id' attribute.", locator);
			
			try
			{
				stationID = Integer.parseInt(stationIDString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException(xcp.getMessage(), locator);
			}
			
			if (!usedIDs.contains(new Integer(stationID)))
				throw new SAXParseException("The connection has got a station assigned that has not been defined before.", locator);
			
			Station station = (Station)idToStationMap.get(new Integer(stationID));
			
			// can not be null if the ID is present
			if (station.GetMeanOfTransportation() != connectionMeanOfTransportation)
				throw new SAXParseException("The station assigned to the connection does not support the mean of transportation of the connection.", locator);
			
			connectionStations.add(station);
		}
		
		// arrival times
		if (name.compareTo("ArrivalTimes") == 0)
		{
			if (!fInConnection)
				throw new SAXParseException("ArrivalTimes tag outside the Connection tag.", locator);
			
			fInArrivalTimes = true;
		}
		
		if (name.compareTo("ArrivalTime") == 0)
		{
			if (!fInArrivalTimes)
				throw new SAXParseException("ArrivalTime tag outside the ArrivalTimes tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("ArrivalTime tag has an invalid number of attributes.", locator);
						
			String arrivalTimeString = attributes.getValue("time");
			
			if (arrivalTimeString == null)
				throw new SAXParseException("ArrivalTime tag has no `time' attribute.", locator);
			
			try
			{
				arrivalTime = TravelTime.ParseTime(arrivalTimeString);
			}
			
			catch (ParseException xcp)
			{
				throw new SAXParseException("Arrival time is incorrect: " + xcp.getMessage(), locator);
			}
			
			if (arrivalTime < 0)
				throw new SAXParseException("Arrival time is incorrect.", locator);
			
			connectionArrivalTimes.add(new Integer(arrivalTime));
		}
		
		// departure times
		if (name.compareTo("DepartureTimes") == 0)
		{
			if (!fInConnection)
				throw new SAXParseException("DepartureTimes tag outside the Connection tag.", locator);
			
			fInDepartureTimes = true;
		}
		
		if (name.compareTo("DepartureTime") == 0)
		{
			if (!fInDepartureTimes)
				throw new SAXParseException("DepartureTime tag outside the DepartureTimes tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("DepartureTime tag has an invalid number of attributes.", locator);
						
			String departureTimeString = attributes.getValue("time");
			
			if (departureTimeString == null)
				throw new SAXParseException("DepartureTime tag has no `time' attribute.", locator);
			
			try
			{
				departureTime = TravelTime.ParseTime(departureTimeString);
			}
			
			catch (ParseException xcp)
			{
				throw new SAXParseException("Departure time is incorrect: " + xcp.getMessage(), locator);
			}
			
			if (departureTime < 0)
				throw new SAXParseException("Departure time is incorrect.", locator);
			
			connectionDepartureTimes.add(new Integer(departureTime));
		}
		
		// day of week
		if (name.compareTo("DaysOfWeek") == 0)
		{
			if (!fInConnection)
				throw new SAXParseException("DaysOfWeek tag outside the Connection tag.", locator);
			
			fInDaysOfWeek = true;
			
			dayOfWeekFlag = 0;
		}
		
		// Monday
		if (name.compareTo("Monday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Monday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Monday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Monday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.MONDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Monday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// Tuesday
		if (name.compareTo("Tuesday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Tuesday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Tuesday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Tuesday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.TUESDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Tuesday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// Wednesday
		if (name.compareTo("Wednesday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Wednesday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Wednesday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Wednesday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.WEDNESDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Wednesday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// Thursday
		if (name.compareTo("Thursday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Thursday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Thursday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Thursday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.THURSDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Thursday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// Friday
		if (name.compareTo("Friday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Friday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Friday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Friday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.FRIDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Friday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// Saturday
		if (name.compareTo("Saturday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Saturday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Saturday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Saturday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.SATURDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Saturday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// Sunday
		if (name.compareTo("Sunday") == 0)
		{
			if (!fInDaysOfWeek)
				throw new SAXParseException("Sunday tag outside the DaysOfWeek tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Sunday tag has an invalid number of attributes.", locator);
						
			String existsString = attributes.getValue("exists");
			
			if (existsString == null)
				throw new SAXParseException("Sunday tag has no `exists' attribute.", locator);
			
			if (existsString.compareTo("true") == 0)
				dayOfWeekFlag |=  Connection.SUNDAY;
			else
			if (existsString.compareTo("false") != 0)
				throw new SAXParseException("Sunday tag's `exists' attribute must be `true' or `false'.", locator);
		}
		
		// location
		if (name.compareTo("Location") == 0)
		{
			if (!fInCity && !fInAirport && !fInTrainStation && !fInBusStop && !fInHarbour)
				throw new SAXParseException("Location tag outside one of the following tags: City, Airport, BusStop, TrainStation and Harbour.", locator);
			
			fInLocation = true;
			
			x = y = -1;
			
			continentName = islandName = null;
		}
		
		// x
		if (name.compareTo("X") == 0)
		{
			if (!fInLocation)
				throw new SAXParseException("X tag outside the Location tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("X tag has an invalid number of attributes.", locator);
						
			String xString = attributes.getValue("x");
			
			if (xString == null)
				throw new SAXParseException("X tag has no `x' attribute.", locator);
			
			try
			{
				x = Integer.parseInt(xString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException("The `x' attribute is not correct: " + xcp.getMessage(), locator);
			}
		}
		
		// y
		if (name.compareTo("Y") == 0)
		{
			if (!fInLocation)
				throw new SAXParseException("Y tag outside the Location tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("Y tag has an invalid number of attributes.", locator);
						
			String yString = attributes.getValue("y");
			
			if (yString == null)
				throw new SAXParseException("Y tag has no `y' attribute.", locator);
			
			try
			{
				y = Integer.parseInt(yString);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new SAXParseException("The `y' attribute is not correct: " + xcp.getMessage(), locator);
			}
		}
		
		// continentName
		if (name.compareTo("ContinentName") == 0)
		{
			if (!fInLocation)
				throw new SAXParseException("ContinentName tag outside the Location tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("ContinentName tag has an invalid number of attributes.", locator);
						
			String nameString = attributes.getValue("name");
			
			if (nameString == null)
				throw new SAXParseException("ContinentName tag has no `name' attribute.", locator);
			
			continentName = nameString;
		}
		
		// island name
		if (name.compareTo("IslandName") == 0)
		{
			if (!fInLocation)
				throw new SAXParseException("IslandName tag outside the Location tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("IslandName tag has an invalid number of attributes.", locator);
						
			String nameString = attributes.getValue("name");
			
			if (nameString == null)
				throw new SAXParseException("IslandName tag has no `name' attribute.", locator);
			
			islandName = nameString;
		}
	}
			
	/**
	 * @author Piotrek
	 * @see org.xml.sax.ContentHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void endElement(String uriNamespace, String qualifiedName, String simpleName) throws SAXParseException
	{
		String name = simpleName;
		
		if (name.compareTo("") == 0) 
			name = qualifiedName; // not namespace aware
		
		// world
		if (name.compareTo("World") == 0)
			fInWorld = false;
		
		// cities
		if (name.compareTo("Cities") == 0)
			fInCities = false;
		
		if (name.compareTo("City") == 0)
		{
			if (!IsLocationParsed())
				throw new SAXParseException("The location tag for the city is missing or is incomplete.", locator);
			
			cities.add(new City(GetLocation(), cityName));
		}
		
		// airports
		if (name.compareTo("Airports") == 0)
			fInAirports = false;
		
		if (name.compareTo("Airport") == 0)
		{
			if (!IsLocationParsed())
				throw new SAXParseException("The location tag for the airport is missing or is incomplete.", locator);
			
			Airport airport = new Airport(GetLocation(), airportName);
			
			stations.add(airport);
			
			idToStationMap.put(new Integer(airportID), airport);
			
			fInAirport = false;
		}
		
		// bus stops
		if (name.compareTo("BusStops") == 0)
			fInBusStops = false;
		
		if (name.compareTo("BusStop") == 0)
		{
			if (!IsLocationParsed())
				throw new SAXParseException("The location tag for the bus stop is missing or is incomplete.", locator);
			
			BusStop busStop = new BusStop(GetLocation(), busStopName);
			
			stations.add(busStop);
			
			idToStationMap.put(new Integer(busStopID), busStop);
			
			fInBusStop = false;
		}
		
		// train stations
		if (name.compareTo("TrainStations") == 0)
			fInTrainStations = false;
		
		if (name.compareTo("TrainStation") == 0)
		{
			if (!IsLocationParsed())
				throw new SAXParseException("The location tag for the train station is missing or is incomplete.", locator);
			
			TrainStation trainStation = new TrainStation(GetLocation(), trainStationName);
			
			stations.add(trainStation);
			
			idToStationMap.put(new Integer(trainStationID), trainStation);
			
			fInTrainStation = false;
		}
		
		// harbours
		if (name.compareTo("Harbours") == 0)
			fInHarbours = false;
		
		if (name.compareTo("Harbour") == 0)
		{
			if (!IsLocationParsed())
				throw new SAXParseException("The location tag for the harbour is missing or is incomplete.", locator);
			
			Harbour harbour = new Harbour(GetLocation(), harbourName);
			
			stations.add(harbour);
			
			idToStationMap.put(new Integer(harbourID), harbour);
			
			fInHarbour = false;
		}
		
		// connections
		if (name.compareTo("Connections") == 0)
			fInConnections = false;
		
		if (name.compareTo("Connection") == 0)
		{
			// sizes
			if (connectionStations.size() != connectionArrivalTimes.size() || connectionStations.size() != connectionDepartureTimes.size())
				throw new SAXParseException("The number of stations for the connection must be the same as numbers of arrival and departure times.", locator);
			
			if (connectionStations.size() < 2)
				throw new SAXParseException("The connection must have at least two stations.", locator);
			
			// times order
			int previousArrivalTime = ((Integer)connectionArrivalTimes.get(0)).intValue();
			int previousDepartureTime = ((Integer)connectionDepartureTimes.get(0)).intValue();
			
			if (previousDepartureTime < previousArrivalTime)
				throw new SAXParseException("The arrival and departure times are not in line.", locator);
			
			for (int i = 1; i < connectionArrivalTimes.size(); ++i)
			{
				int currentArrivalTime = ((Integer)connectionArrivalTimes.get(i)).intValue();
				int currentDepartureTime = ((Integer)connectionDepartureTimes.get(i)).intValue();
								
				if (currentArrivalTime < previousArrivalTime)
					throw new SAXParseException("The arrival times order is incorrect.", locator);
				
				if (currentDepartureTime < previousDepartureTime)
					throw new SAXParseException("The departure times order is incorrect.", locator);
				
				previousArrivalTime = currentArrivalTime;
				previousDepartureTime = currentDepartureTime;
				
				if (previousDepartureTime < previousArrivalTime)
					throw new SAXParseException("The arrival and departure times are not in line.", locator);
			}
			
			// creating a connection
			Connection connection = new Connection(connectionName, connectionMeanOfTransportation, connectionStations, connectionArrivalTimes, connectionDepartureTimes, dayOfWeekFlag);
			
			connections.add(connection);
		}
		
		if (name.compareTo("StationIDs") == 0)
			fInStationIDs = false;
		
		if (name.compareTo("ArrivalTimes") == 0)
			fInArrivalTimes = false;
			
		if (name.compareTo("DepartureTimes") == 0)
			fInDepartureTimes = false;
		
		if (name.compareTo("DaysOfWeek") == 0)
		{
			fInDaysOfWeek = false;
			
			if (dayOfWeekFlag == 0)
				throw new SAXParseException("There are no day of week flags defined for the connection or all of them are set to false.", locator);
		}
		
		if (name.compareTo("Location") == 0)
			fInLocation = false;
	}
		
	/**
	 * @author Piotrek
	 * @see org.xml.sax.ContentHandler#setDocumentLocator(org.xml.sax.Locator)
	 */
	public void setDocumentLocator(Locator locator)
	{
		this.locator = locator;
	}
	
	/**
	 * The SAX locator.
	 * @author Piotrek
	 */
	private Locator locator;
	
	/**
	 * Changes the world to the parsed one.
	 * @author Piotrek
	 * @throws Exception Thrown in case of data error.
	 */
	private void ChangeWorld() throws Exception
	{
		World.Clear();
		
		World world = World.GetInstance();
		
		// cities
		Iterator iterator = cities.iterator();
		
		while (iterator.hasNext())
			world.AddCity((City)iterator.next());
		
		// stations
		
		iterator = stations.iterator();
		
		while (iterator.hasNext())
		{
			Station station = (Station)iterator.next();
			
			switch (station.GetMeanOfTransportation())
			{
				case MeanOfTransportation.AIRPLANE:
					
					world.AddAirport((Airport)station);
				
					break;
					
				case MeanOfTransportation.BUS:
					
					world.AddBusStop((BusStop)station);
				
					break;
					
				case MeanOfTransportation.TRAIN:
					
					world.AddTrainStation((TrainStation)station);
				
					break;
					
				case MeanOfTransportation.SHIP:
					
					world.AddHarbour((Harbour)station);
				
					break;
					
				default:
					
					throw new Exception("Unexpected mean of transportation in WorldFileParser.ChangeWorld().");
			}
		}
		
		// connections
		
		iterator = connections.iterator();
		
		while (iterator.hasNext())
			world.AddConnection((Connection)iterator.next());
	}
	
	/**
	 * Indicates whether the location tag has been parsed completely.
	 * @author Piotrek
	 * @author Piotrek
	 * @return A boolean value indicating whether the location tag has been parsed completely.
	 */
	private boolean IsLocationParsed()
	{
		if (x == -1)
			return false;
		
		if (y == -1)
			return false;
		
		if (continentName == null)
			return false;
		
		if (islandName == null)
			return false;
		
		return true;
	}
	
	/**
	 * Gets the location built basing on the parsed data.
	 * @author Piotrek
	 * @return The location builr basing on the parsed data.
	 */
	private Location GetLocation()
	{
		Location location = new Location(x, y);
		
		location.continentName = continentName;
		location.islandName = islandName;
		
		return location;
	}
	
	/**
	 * Clears the data of the connection. Done to detect missing location tags.
	 * @author Piotrek
	 */
	private void ClearLocation()
	{
		x = y = -1;
		
		continentName = islandName = null;
	}
}
